// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: proto/nodeinstance/create/v1/creation.proto
// Protobuf C++ Version: 7.35.1

#include "proto/nodeinstance/create/v1/creation.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::nodeinstance::create::v1::CreateNodeInstance
        {&::_pbi::kDescriptorMethods, &::descriptor_table_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto, /* tracker*/ nullptr,},
        // ::nodeinstance::create::v1::CreateNodeInstanceResult
        {&::_pbi::kDescriptorMethods, &::descriptor_table_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace nodeinstance {
namespace create {
namespace v1 {
class CreateNodeInstanceResult::_Internal {
 public:
  using HasBits = decltype(::std::declval<CreateNodeInstanceResult>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CreateNodeInstanceResult, _impl_._has_bits_);
};

constexpr CreateNodeInstanceResult::ParseTableT_ CreateNodeInstanceResult::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CreateNodeInstanceResult, _impl_._has_bits_),
      0, // no _extensions_
      2, 8,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967292,  // skipmap
      offsetof(ParseTableT_, field_entries),
      2,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::nodeinstance::create::v1::CreateNodeInstanceResult>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // string machine_guid = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CreateNodeInstanceResult, _impl_.machine_guid_)}},
      // string node_id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CreateNodeInstanceResult, _impl_.node_id_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string node_id = 1;
      {PROTOBUF_FIELD_OFFSET(CreateNodeInstanceResult, _impl_.node_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string machine_guid = 2;
      {PROTOBUF_FIELD_OFFSET(CreateNodeInstanceResult, _impl_.machine_guid_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\57\7\14\0\0\0\0\0"
      "nodeinstance.create.v1.CreateNodeInstanceResult"
      "node_id"
      "machine_guid"
    }},
  };
}


inline constexpr CreateNodeInstanceResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        node_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        machine_guid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
constexpr CreateNodeInstanceResult::CreateNodeInstanceResult(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CreateNodeInstanceResult::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CreateNodeInstanceResult(arena);
}
constexpr auto CreateNodeInstanceResult::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CreateNodeInstanceResult), alignof(CreateNodeInstanceResult));
}
constexpr auto CreateNodeInstanceResult::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CreateNodeInstanceResult::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CreateNodeInstanceResult>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CreateNodeInstanceResult::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CreateNodeInstanceResult>(), &CreateNodeInstanceResult::ByteSizeLong,
              &CreateNodeInstanceResult::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CreateNodeInstanceResult, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CreateNodeInstanceResultGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CreateNodeInstanceResultGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CreateNodeInstanceResult_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CreateNodeInstanceResult::InternalGenerateClassData_(
            _default, &CreateNodeInstanceResult_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CreateNodeInstanceResult>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CreateNodeInstanceResultGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CreateNodeInstanceResult _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CreateNodeInstanceResult>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CreateNodeInstanceResultGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CreateNodeInstanceResultGlobalsTypeInternal CreateNodeInstanceResult_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CreateNodeInstanceResult_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CreateNodeInstanceResult_globals_.GetClassData();
#else
  return CreateNodeInstanceResult_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CreateNodeInstance::_Internal {
 public:
  using HasBits = decltype(::std::declval<CreateNodeInstance>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_._has_bits_);
};

constexpr CreateNodeInstance::ParseTableT_ CreateNodeInstance::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_._has_bits_),
      0, // no _extensions_
      6, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967232,  // skipmap
      offsetof(ParseTableT_, field_entries),
      6,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::nodeinstance::create::v1::CreateNodeInstance>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string claim_id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.claim_id_)}},
      // string machine_guid = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.machine_guid_)}},
      // string hostname = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.hostname_)}},
      // int32 vmq_instance_id = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(CreateNodeInstance, _impl_.vmq_instance_id_), 4>(),
       {32, 4, 0,
        PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.vmq_instance_id_)}},
      // string mqtt_broker_addr = 5;
      {::_pbi::TcParser::FastUS1,
       {42, 3, 0,
        PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.mqtt_broker_addr_)}},
      // int32 hops = 6;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(CreateNodeInstance, _impl_.hops_), 5>(),
       {48, 5, 0,
        PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.hops_)}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // string claim_id = 1;
      {PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.claim_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string machine_guid = 2;
      {PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.machine_guid_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string hostname = 3;
      {PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.hostname_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // int32 vmq_instance_id = 4;
      {PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.vmq_instance_id_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // string mqtt_broker_addr = 5;
      {PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.mqtt_broker_addr_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // int32 hops = 6;
      {PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.hops_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
    }},
    // no aux_entries
    {{
      "\51\10\14\10\0\20\0\0"
      "nodeinstance.create.v1.CreateNodeInstance"
      "claim_id"
      "machine_guid"
      "hostname"
      "mqtt_broker_addr"
    }},
  };
}


inline constexpr CreateNodeInstance::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        claim_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        machine_guid_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        hostname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        mqtt_broker_addr_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        vmq_instance_id_{0},
        hops_{0} {}

template <typename>
constexpr CreateNodeInstance::CreateNodeInstance(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CreateNodeInstance::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CreateNodeInstance(arena);
}
constexpr auto CreateNodeInstance::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CreateNodeInstance), alignof(CreateNodeInstance));
}
constexpr auto CreateNodeInstance::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CreateNodeInstance::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CreateNodeInstance>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CreateNodeInstance::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CreateNodeInstance>(), &CreateNodeInstance::ByteSizeLong,
              &CreateNodeInstance::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CreateNodeInstanceGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CreateNodeInstanceGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CreateNodeInstance_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CreateNodeInstance::InternalGenerateClassData_(
            _default, &CreateNodeInstance_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CreateNodeInstance>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CreateNodeInstanceGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CreateNodeInstance _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CreateNodeInstance>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CreateNodeInstanceGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CreateNodeInstanceGlobalsTypeInternal CreateNodeInstance_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CreateNodeInstance_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CreateNodeInstance_globals_.GetClassData();
#else
  return CreateNodeInstance_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace v1
}  // namespace create
}  // namespace nodeinstance
static constexpr const ::_pb::EnumDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_enum_descriptors_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto = nullptr;
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto = nullptr;
const ::uint32_t
    TableStruct_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstance, _impl_._has_bits_),
        9, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstance, _impl_.claim_id_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstance, _impl_.machine_guid_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstance, _impl_.hostname_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstance, _impl_.vmq_instance_id_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstance, _impl_.mqtt_broker_addr_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstance, _impl_.hops_),
        0,
        1,
        2,
        4,
        3,
        5,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstanceResult, _impl_._has_bits_),
        5, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstanceResult, _impl_.node_id_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::create::v1::CreateNodeInstanceResult, _impl_.machine_guid_),
        0,
        1,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::nodeinstance::create::v1::CreateNodeInstance)},
        {15, sizeof(::nodeinstance::create::v1::CreateNodeInstanceResult)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::nodeinstance::create::v1::CreateNodeInstance_globals_,
        &::nodeinstance::create::v1::CreateNodeInstanceResult_globals_,
};
const char descriptor_table_protodef_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n+proto/nodeinstance/create/v1/creation."
    "proto\022\026nodeinstance.create.v1\"\217\001\n\022Create"
    "NodeInstance\022\020\n\010claim_id\030\001 \001(\t\022\024\n\014machin"
    "e_guid\030\002 \001(\t\022\020\n\010hostname\030\003 \001(\t\022\027\n\017vmq_in"
    "stance_id\030\004 \001(\005\022\030\n\020mqtt_broker_addr\030\005 \001("
    "\t\022\014\n\004hops\030\006 \001(\005\"A\n\030CreateNodeInstanceRes"
    "ult\022\017\n\007node_id\030\001 \001(\t\022\024\n\014machine_guid\030\002 \001"
    "(\tB1Z/node_instance/creation/v1;node_ins"
    "tancecreationb\006proto3"
};
static ::absl::once_flag descriptor_table_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto = {
    false,
    false,
    341,
    descriptor_table_protodef_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto,
    "proto/nodeinstance/create/v1/creation.proto",
    &descriptor_table_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto_once,
    nullptr,
    0,
    2,
    schemas,
    file_message_globals,
    TableStruct_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto::offsets,
    file_level_enum_descriptors_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto,
    file_level_service_descriptors_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto,
};
namespace nodeinstance {
namespace create {
namespace v1 {
// ===================================================================

CreateNodeInstance::CreateNodeInstance(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CreateNodeInstance_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:nodeinstance.create.v1.CreateNodeInstance)
}
PROTOBUF_NDEBUG_INLINE CreateNodeInstance::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::nodeinstance::create::v1::CreateNodeInstance& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        claim_id_(arena, from.claim_id_),
        machine_guid_(arena, from.machine_guid_),
        hostname_(arena, from.hostname_),
        mqtt_broker_addr_(arena, from.mqtt_broker_addr_) {}

CreateNodeInstance::CreateNodeInstance(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CreateNodeInstance& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CreateNodeInstance_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CreateNodeInstance* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, vmq_instance_id_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, vmq_instance_id_),
           offsetof(Impl_, hops_) -
               offsetof(Impl_, vmq_instance_id_) +
               sizeof(Impl_::hops_));

  // @@protoc_insertion_point(copy_constructor:nodeinstance.create.v1.CreateNodeInstance)
}
PROTOBUF_NDEBUG_INLINE CreateNodeInstance::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        claim_id_(arena),
        machine_guid_(arena),
        hostname_(arena),
        mqtt_broker_addr_(arena) {}

inline void CreateNodeInstance::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, vmq_instance_id_),
           0,
           offsetof(Impl_, hops_) -
               offsetof(Impl_, vmq_instance_id_) +
               sizeof(Impl_::hops_));
}
CreateNodeInstance::~CreateNodeInstance() {
  // @@protoc_insertion_point(destructor:nodeinstance.create.v1.CreateNodeInstance)
  SharedDtor(*this);
}
inline void CreateNodeInstance::SharedDtor(MessageLite& self) {
  CreateNodeInstance& this_ = static_cast<CreateNodeInstance&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.claim_id_.Destroy();
  this_._impl_.machine_guid_.Destroy();
  this_._impl_.hostname_.Destroy();
  this_._impl_.mqtt_broker_addr_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CreateNodeInstance_class_data_ =
        CreateNodeInstance::InternalGenerateClassData_(CreateNodeInstance_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CreateNodeInstance::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CreateNodeInstance_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CreateNodeInstance_class_data_.tc_table);
  return CreateNodeInstance_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CreateNodeInstance::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CreateNodeInstance_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CreateNodeInstance_globals_));
  return CreateNodeInstance_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CreateNodeInstance::ParseTableT_
    CreateNodeInstance::_table_ =
        CreateNodeInstance::InternalGenerateParseTable_(CreateNodeInstance_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CreateNodeInstance::Clear() {
// @@protoc_insertion_point(message_clear_start:nodeinstance.create.v1.CreateNodeInstance)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.claim_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.machine_guid_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.hostname_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.mqtt_broker_addr_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000030U)) {
    ::memset(&_impl_.vmq_instance_id_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.hops_) -
        reinterpret_cast<char*>(&_impl_.vmq_instance_id_)) + sizeof(_impl_.hops_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CreateNodeInstance::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CreateNodeInstance& this_ = static_cast<const CreateNodeInstance&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CreateNodeInstance::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CreateNodeInstance& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:nodeinstance.create.v1.CreateNodeInstance)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string claim_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_claim_id().empty()) {
      const ::std::string& _s = this_._internal_claim_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.create.v1.CreateNodeInstance.claim_id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string machine_guid = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_machine_guid().empty()) {
      const ::std::string& _s = this_._internal_machine_guid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.create.v1.CreateNodeInstance.machine_guid");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // string hostname = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_hostname().empty()) {
      const ::std::string& _s = this_._internal_hostname();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.create.v1.CreateNodeInstance.hostname");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  // int32 vmq_instance_id = 4;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (this_._internal_vmq_instance_id() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<4>(
              stream, this_._internal_vmq_instance_id(), target);
    }
  }

  // string mqtt_broker_addr = 5;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (!this_._internal_mqtt_broker_addr().empty()) {
      const ::std::string& _s = this_._internal_mqtt_broker_addr();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.create.v1.CreateNodeInstance.mqtt_broker_addr");
      target = stream->WriteStringMaybeAliased(5, _s, target);
    }
  }

  // int32 hops = 6;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (this_._internal_hops() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<6>(
              stream, this_._internal_hops(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:nodeinstance.create.v1.CreateNodeInstance)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CreateNodeInstance::ByteSizeLong(const MessageLite& base) {
  const CreateNodeInstance& this_ = static_cast<const CreateNodeInstance&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CreateNodeInstance::ByteSizeLong() const {
  const CreateNodeInstance& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:nodeinstance.create.v1.CreateNodeInstance)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    // string claim_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_claim_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_claim_id());
      }
    }
    // string machine_guid = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_machine_guid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_machine_guid());
      }
    }
    // string hostname = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_hostname().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_hostname());
      }
    }
    // string mqtt_broker_addr = 5;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!this_._internal_mqtt_broker_addr().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mqtt_broker_addr());
      }
    }
    // int32 vmq_instance_id = 4;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (this_._internal_vmq_instance_id() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_vmq_instance_id());
      }
    }
    // int32 hops = 6;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (this_._internal_hops() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_hops());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CreateNodeInstance::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CreateNodeInstance*>(&to_msg);
  auto& from = static_cast<const CreateNodeInstance&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:nodeinstance.create.v1.CreateNodeInstance)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_claim_id().empty()) {
        _this->_internal_set_claim_id(from._internal_claim_id());
      } else {
        if (_this->_impl_.claim_id_.IsDefault()) {
          _this->_internal_set_claim_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_machine_guid().empty()) {
        _this->_internal_set_machine_guid(from._internal_machine_guid());
      } else {
        if (_this->_impl_.machine_guid_.IsDefault()) {
          _this->_internal_set_machine_guid("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_hostname().empty()) {
        _this->_internal_set_hostname(from._internal_hostname());
      } else {
        if (_this->_impl_.hostname_.IsDefault()) {
          _this->_internal_set_hostname("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!from._internal_mqtt_broker_addr().empty()) {
        _this->_internal_set_mqtt_broker_addr(from._internal_mqtt_broker_addr());
      } else {
        if (_this->_impl_.mqtt_broker_addr_.IsDefault()) {
          _this->_internal_set_mqtt_broker_addr("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (from._internal_vmq_instance_id() != 0) {
        _this->_impl_.vmq_instance_id_ = from._impl_.vmq_instance_id_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (from._internal_hops() != 0) {
        _this->_impl_.hops_ = from._impl_.hops_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CreateNodeInstance::CopyFrom(const CreateNodeInstance& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:nodeinstance.create.v1.CreateNodeInstance)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CreateNodeInstance::InternalSwap(CreateNodeInstance* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.claim_id_, &other->_impl_.claim_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.machine_guid_, &other->_impl_.machine_guid_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.hostname_, &other->_impl_.hostname_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mqtt_broker_addr_, &other->_impl_.mqtt_broker_addr_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.hops_)
      + sizeof(CreateNodeInstance::_impl_.hops_)
      - PROTOBUF_FIELD_OFFSET(CreateNodeInstance, _impl_.vmq_instance_id_)>(
          reinterpret_cast<char*>(&_impl_.vmq_instance_id_),
          reinterpret_cast<char*>(&other->_impl_.vmq_instance_id_));
}

::google::protobuf::Metadata CreateNodeInstance::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

CreateNodeInstanceResult::CreateNodeInstanceResult(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CreateNodeInstanceResult_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:nodeinstance.create.v1.CreateNodeInstanceResult)
}
PROTOBUF_NDEBUG_INLINE CreateNodeInstanceResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::nodeinstance::create::v1::CreateNodeInstanceResult& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        node_id_(arena, from.node_id_),
        machine_guid_(arena, from.machine_guid_) {}

CreateNodeInstanceResult::CreateNodeInstanceResult(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CreateNodeInstanceResult& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CreateNodeInstanceResult_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CreateNodeInstanceResult* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:nodeinstance.create.v1.CreateNodeInstanceResult)
}
PROTOBUF_NDEBUG_INLINE CreateNodeInstanceResult::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        node_id_(arena),
        machine_guid_(arena) {}

inline void CreateNodeInstanceResult::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
CreateNodeInstanceResult::~CreateNodeInstanceResult() {
  // @@protoc_insertion_point(destructor:nodeinstance.create.v1.CreateNodeInstanceResult)
  SharedDtor(*this);
}
inline void CreateNodeInstanceResult::SharedDtor(MessageLite& self) {
  CreateNodeInstanceResult& this_ = static_cast<CreateNodeInstanceResult&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.node_id_.Destroy();
  this_._impl_.machine_guid_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CreateNodeInstanceResult_class_data_ =
        CreateNodeInstanceResult::InternalGenerateClassData_(CreateNodeInstanceResult_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CreateNodeInstanceResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CreateNodeInstanceResult_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CreateNodeInstanceResult_class_data_.tc_table);
  return CreateNodeInstanceResult_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CreateNodeInstanceResult::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CreateNodeInstanceResult_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CreateNodeInstanceResult_globals_));
  return CreateNodeInstanceResult_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CreateNodeInstanceResult::ParseTableT_
    CreateNodeInstanceResult::_table_ =
        CreateNodeInstanceResult::InternalGenerateParseTable_(CreateNodeInstanceResult_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CreateNodeInstanceResult::Clear() {
// @@protoc_insertion_point(message_clear_start:nodeinstance.create.v1.CreateNodeInstanceResult)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.node_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.machine_guid_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CreateNodeInstanceResult::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CreateNodeInstanceResult& this_ = static_cast<const CreateNodeInstanceResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CreateNodeInstanceResult::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CreateNodeInstanceResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:nodeinstance.create.v1.CreateNodeInstanceResult)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string node_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_node_id().empty()) {
      const ::std::string& _s = this_._internal_node_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.create.v1.CreateNodeInstanceResult.node_id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string machine_guid = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_machine_guid().empty()) {
      const ::std::string& _s = this_._internal_machine_guid();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.create.v1.CreateNodeInstanceResult.machine_guid");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:nodeinstance.create.v1.CreateNodeInstanceResult)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CreateNodeInstanceResult::ByteSizeLong(const MessageLite& base) {
  const CreateNodeInstanceResult& this_ = static_cast<const CreateNodeInstanceResult&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CreateNodeInstanceResult::ByteSizeLong() const {
  const CreateNodeInstanceResult& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:nodeinstance.create.v1.CreateNodeInstanceResult)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    // string node_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_node_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_node_id());
      }
    }
    // string machine_guid = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_machine_guid().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_machine_guid());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CreateNodeInstanceResult::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CreateNodeInstanceResult*>(&to_msg);
  auto& from = static_cast<const CreateNodeInstanceResult&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:nodeinstance.create.v1.CreateNodeInstanceResult)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_node_id().empty()) {
        _this->_internal_set_node_id(from._internal_node_id());
      } else {
        if (_this->_impl_.node_id_.IsDefault()) {
          _this->_internal_set_node_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_machine_guid().empty()) {
        _this->_internal_set_machine_guid(from._internal_machine_guid());
      } else {
        if (_this->_impl_.machine_guid_.IsDefault()) {
          _this->_internal_set_machine_guid("");
        }
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CreateNodeInstanceResult::CopyFrom(const CreateNodeInstanceResult& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:nodeinstance.create.v1.CreateNodeInstanceResult)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CreateNodeInstanceResult::InternalSwap(CreateNodeInstanceResult* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.node_id_, &other->_impl_.node_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.machine_guid_, &other->_impl_.machine_guid_, arena);
}

::google::protobuf::Metadata CreateNodeInstanceResult::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace create
}  // namespace nodeinstance
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_proto_2fnodeinstance_2fcreate_2fv1_2fcreation_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
