// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef NETDATA_MMAP_LIMIT_H
#define NETDATA_MMAP_LIMIT_H

#include "libnetdata/common.h"

unsigned long long os_mmap_limit(void);

#endif //NETDATA_MMAP_LIMIT_H
