// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef NETDATA_DAEMON_SYSTEMD_WATCHER_H
#define NETDATA_DAEMON_SYSTEMD_WATCHER_H

void systemd_watcher_thread(void *arg);

#endif //NETDATA_DAEMON_SYSTEMD_WATCHER_H
