/* This file is automatically generated by Lemon from input grammar
** source file "parser.y".
*/
/*
** 2000-05-29
**
** The author disclaims copyright to this source code.  In place of
** a legal notice, here is a blessing:
**
**    May you do good and not evil.
**    May you find forgiveness for yourself and forgive others.
**    May you share freely, never taking more than you give.
**
*************************************************************************
** Driver template for the LEMON parser generator.
**
** The "lemon" program processes an LALR(1) input grammar file, then uses
** this template to construct a parser.  The "lemon" program inserts text
** at each "%%" line.  Also, any "P-a-r-s-e" identifier prefix (without the
** interstitial "-" characters) contained in this template is changed into
** the value of the %name directive from the grammar.  Otherwise, the content
** of this template is copied straight through into the generate parser
** source file.
**
** The following is the concatenation of all %include directives from the
** input grammar file:
*/
/************ Begin %include sections from the grammar ************************/
#include "../eval-internal.h"
#include "parser_internal.h"
#include <assert.h>
#line 33 "parser.c"
/**************** End of %include directives **********************************/
/* These constants specify the various numeric values for terminal symbols.
***************** Begin token definitions *************************************/
#ifndef TOK_NUMBER
#define TOK_NUMBER                          1
#define TOK_VARIABLE                        2
#define TOK_LPAREN                          3
#define TOK_RPAREN                          4
#define TOK_PLUS                            5
#define TOK_UPLUS                           6
#define TOK_MINUS                           7
#define TOK_UMINUS                          8
#define TOK_NOT                             9
#define TOK_FUNCTION_ABS                   10
#define TOK_MULTIPLY                       11
#define TOK_DIVIDE                         12
#define TOK_MODULO                         13
#define TOK_AND                            14
#define TOK_OR                             15
#define TOK_EQ                             16
#define TOK_NE                             17
#define TOK_LT                             18
#define TOK_LE                             19
#define TOK_GT                             20
#define TOK_GE                             21
#define TOK_QMARK                          22
#define TOK_COLON                          23
#endif
/**************** End token definitions ***************************************/

/* The next sections is a series of control #defines.
** various aspects of the generated parser.
**    YYCODETYPE         is the data type used to store the integer codes
**                       that represent terminal and non-terminal symbols.
**                       "unsigned char" is used if there are fewer than
**                       256 symbols.  Larger types otherwise.
**    YYNOCODE           is a number of type YYCODETYPE that is not used for
**                       any terminal or nonterminal symbol.
**    YYFALLBACK         If defined, this indicates that one or more tokens
**                       (also known as: "terminal symbols") have fall-back
**                       values which should be used if the original symbol
**                       would not parse.  This permits keywords to sometimes
**                       be used as identifiers, for example.
**    YYACTIONTYPE       is the data type used for "action codes" - numbers
**                       that indicate what to do in response to the next
**                       token.
**    ParseTOKENTYPE     is the data type used for minor type for terminal
**                       symbols.  Background: A "minor type" is a semantic
**                       value associated with a terminal or non-terminal
**                       symbols.  For example, for an "ID" terminal symbol,
**                       the minor type might be the name of the identifier.
**                       Each non-terminal can have a different minor type.
**                       Terminal symbols all have the same minor type, though.
**                       This macros defines the minor type for terminal 
**                       symbols.
**    YYMINORTYPE        is the data type used for all minor types.
**                       This is typically a union of many types, one of
**                       which is ParseTOKENTYPE.  The entry in the union
**                       for terminal symbols is called "yy0".
**    YYSTACKDEPTH       is the maximum depth of the parser's stack.  If
**                       zero the stack is dynamically sized using realloc()
**    ParseARG_SDECL     A static variable declaration for the %extra_argument
**    ParseARG_PDECL     A parameter declaration for the %extra_argument
**    ParseARG_PARAM     Code to pass %extra_argument as a subroutine parameter
**    ParseARG_STORE     Code to store %extra_argument into yypParser
**    ParseARG_FETCH     Code to extract %extra_argument from yypParser
**    ParseCTX_*         As ParseARG_ except for %extra_context
**    YYREALLOC          Name of the realloc() function to use
**    YYFREE             Name of the free() function to use
**    YYDYNSTACK         True if stack space should be extended on heap
**    YYERRORSYMBOL      is the code number of the error symbol.  If not
**                       defined, then do no error processing.
**    YYNSTATE           the combined number of states.
**    YYNRULE            the number of rules in the grammar
**    YYNTOKEN           Number of terminal symbols
**    YY_MAX_SHIFT       Maximum value for shift actions
**    YY_MIN_SHIFTREDUCE Minimum value for shift-reduce actions
**    YY_MAX_SHIFTREDUCE Maximum value for shift-reduce actions
**    YY_ERROR_ACTION    The yy_action[] code for syntax error
**    YY_ACCEPT_ACTION   The yy_action[] code for accept
**    YY_NO_ACTION       The yy_action[] code for no-op
**    YY_MIN_REDUCE      Minimum value for reduce actions
**    YY_MAX_REDUCE      Maximum value for reduce actions
**    YY_MIN_DSTRCTR     Minimum symbol value that has a destructor
**    YY_MAX_DSTRCTR     Maximum symbol value that has a destructor
*/
#ifndef INTERFACE
# define INTERFACE 1
#endif
/************* Begin control #defines *****************************************/
#define YYCODETYPE unsigned char
#define YYNOCODE 26
#define YYACTIONTYPE unsigned char
#define ParseTOKENTYPE YYSTYPE
typedef union {
  int yyinit;
  ParseTOKENTYPE yy0;
  EVAL_NODE* yy48;
} YYMINORTYPE;
#ifndef YYSTACKDEPTH
#define YYSTACKDEPTH 100
#endif
#define ParseARG_SDECL EVAL_NODE **result;
#define ParseARG_PDECL ,EVAL_NODE **result
#define ParseARG_PARAM ,result
#define ParseARG_FETCH EVAL_NODE **result=yypParser->result;
#define ParseARG_STORE yypParser->result=result;
#define YYREALLOC realloc
#define YYFREE free
#define YYDYNSTACK 0
#define ParseCTX_SDECL
#define ParseCTX_PDECL
#define ParseCTX_PARAM
#define ParseCTX_FETCH
#define ParseCTX_STORE
#define YYNSTATE             37
#define YYNRULE              22
#define YYNRULE_WITH_ACTION  22
#define YYNTOKEN             24
#define YY_MAX_SHIFT         36
#define YY_MIN_SHIFTREDUCE   47
#define YY_MAX_SHIFTREDUCE   68
#define YY_ERROR_ACTION      69
#define YY_ACCEPT_ACTION     70
#define YY_NO_ACTION         71
#define YY_MIN_REDUCE        72
#define YY_MAX_REDUCE        93
#define YY_MIN_DSTRCTR       24
#define YY_MAX_DSTRCTR       24
/************* End control #defines *******************************************/
#define YY_NLOOKAHEAD ((int)(sizeof(yy_lookahead)/sizeof(yy_lookahead[0])))

/* Define the yytestcase() macro to be a no-op if is not already defined
** otherwise.
**
** Applications can choose to define yytestcase() in the %include section
** to a macro that can assist in verifying code coverage.  For production
** code the yytestcase() macro should be turned off.  But it is useful
** for testing.
*/
#ifndef yytestcase
# define yytestcase(X)
#endif

/* Macro to determine if stack space has the ability to grow using
** heap memory.
*/
#if YYSTACKDEPTH<=0 || YYDYNSTACK
# define YYGROWABLESTACK 1
#else
# define YYGROWABLESTACK 0
#endif

/* Guarantee a minimum number of initial stack slots.
*/
#if YYSTACKDEPTH<=0
# undef YYSTACKDEPTH
# define YYSTACKDEPTH 2  /* Need a minimum stack size */
#endif


/* Next are the tables used to determine what action to take based on the
** current state and lookahead token.  These tables are used to implement
** functions that take a state number and lookahead value and return an
** action integer.  
**
** Suppose the action integer is N.  Then the action is determined as
** follows
**
**   0 <= N <= YY_MAX_SHIFT             Shift N.  That is, push the lookahead
**                                      token onto the stack and goto state N.
**
**   N between YY_MIN_SHIFTREDUCE       Shift to an arbitrary state then
**     and YY_MAX_SHIFTREDUCE           reduce by rule N-YY_MIN_SHIFTREDUCE.
**
**   N == YY_ERROR_ACTION               A syntax error has occurred.
**
**   N == YY_ACCEPT_ACTION              The parser accepts its input.
**
**   N == YY_NO_ACTION                  No such action.  Denotes unused
**                                      slots in the yy_action[] table.
**
**   N between YY_MIN_REDUCE            Reduce by rule N-YY_MIN_REDUCE
**     and YY_MAX_REDUCE
**
** The action table is constructed as a single large table named yy_action[].
** Given state S and lookahead X, the action is computed as either:
**
**    (A)   N = yy_action[ yy_shift_ofst[S] + X ]
**    (B)   N = yy_default[S]
**
** The (A) formula is preferred.  The B formula is used instead if
** yy_lookahead[yy_shift_ofst[S]+X] is not equal to X.
**
** The formulas above are for computing the action when the lookahead is
** a terminal symbol.  If the lookahead is a non-terminal (as occurs after
** a reduce action) then the yy_reduce_ofst[] array is used in place of
** the yy_shift_ofst[] array.
**
** The following are the tables generated in this section:
**
**  yy_action[]        A single table containing all actions.
**  yy_lookahead[]     A table containing the lookahead for each entry in
**                     yy_action.  Used to detect hash collisions.
**  yy_shift_ofst[]    For each state, the offset into yy_action for
**                     shifting terminals.
**  yy_reduce_ofst[]   For each state, the offset into yy_action for
**                     shifting non-terminals after a reduce.
**  yy_default[]       Default action for each state.
**
*********** Begin parsing tables **********************************************/
#define YY_ACTTAB_COUNT (147)
static const YYACTIONTYPE yy_action[] = {
 /*     0 */    24,   70,   25,   21,   30,   20,   31,   18,   32,   33,
 /*    10 */    28,   16,   14,   12,   10,    9,    8,    7,    6,    5,
 /*    20 */     4,    3,    2,    1,   54,   20,   29,   18,   16,   14,
 /*    30 */    12,   16,   14,   12,   10,    9,    8,    7,    6,    5,
 /*    40 */     4,    3,    2,   50,   20,   26,   18,   27,   22,   84,
 /*    50 */    16,   14,   12,   10,    9,    8,    7,    6,    5,    4,
 /*    60 */     3,    2,   72,   11,   20,   78,   18,   20,   83,   18,
 /*    70 */    16,   14,   12,   16,   14,   12,   10,    9,    8,    7,
 /*    80 */     6,    5,    4,    3,    2,   20,   71,   18,   71,   77,
 /*    90 */    82,   16,   14,   12,   10,    9,    8,    7,    6,    5,
 /*   100 */     4,    3,    2,   20,   71,   18,   76,   35,   23,   16,
 /*   110 */    14,   12,   34,   71,    8,    7,    6,    5,    4,    3,
 /*   120 */    20,   71,   18,   71,   71,   71,   16,   14,   12,   71,
 /*   130 */    71,   71,   71,    6,    5,    4,    3,   48,   49,   19,
 /*   140 */    71,   17,   71,   15,   71,   13,   36,
};
static const YYCODETYPE yy_lookahead[] = {
 /*     0 */    24,   25,   24,   24,   24,    5,   24,    7,   24,   24,
 /*    10 */    24,   11,   12,   13,   14,   15,   16,   17,   18,   19,
 /*    20 */    20,   21,   22,   23,    4,    5,   24,    7,   11,   12,
 /*    30 */    13,   11,   12,   13,   14,   15,   16,   17,   18,   19,
 /*    40 */    20,   21,   22,    4,    5,   24,    7,   24,   24,   24,
 /*    50 */    11,   12,   13,   14,   15,   16,   17,   18,   19,   20,
 /*    60 */    21,   22,    0,    3,    5,   24,    7,    5,   24,    7,
 /*    70 */    11,   12,   13,   11,   12,   13,   14,   15,   16,   17,
 /*    80 */    18,   19,   20,   21,   22,    5,   26,    7,   26,   24,
 /*    90 */    24,   11,   12,   13,   14,   15,   16,   17,   18,   19,
 /*   100 */    20,   21,   22,    5,   26,    7,   24,   24,   24,   11,
 /*   110 */    12,   13,   24,   26,   16,   17,   18,   19,   20,   21,
 /*   120 */     5,   26,    7,   26,   26,   26,   11,   12,   13,   26,
 /*   130 */    26,   26,   26,   18,   19,   20,   21,    1,    2,    3,
 /*   140 */    26,    5,   26,    7,   26,    9,   10,   26,   26,   26,
 /*   150 */    26,   26,   26,   26,   26,   26,   26,   26,   26,   26,
 /*   160 */    26,   26,   24,   24,   24,   24,   24,   24,   24,   24,
 /*   170 */    24,
};
#define YY_SHIFT_COUNT    (36)
#define YY_SHIFT_MIN      (0)
#define YY_SHIFT_MAX      (136)
static const unsigned char yy_shift_ofst[] = {
 /*     0 */   136,  136,  136,  136,  136,  136,  136,  136,  136,  136,
 /*    10 */   136,  136,  136,  136,  136,  136,  136,  136,  136,  136,
 /*    20 */   136,    0,   20,   39,   62,   80,   98,   98,  115,  115,
 /*    30 */    59,   59,   59,   59,   17,   17,   60,
};
#define YY_REDUCE_COUNT (20)
#define YY_REDUCE_MIN   (-24)
#define YY_REDUCE_MAX   (88)
static const signed char yy_reduce_ofst[] = {
 /*     0 */   -24,  -22,  -21,  -20,  -18,  -16,  -15,  -14,    2,   21,
 /*    10 */    23,   24,   25,   41,   44,   65,   66,   82,   83,   84,
 /*    20 */    88,
};
static const YYACTIONTYPE yy_default[] = {
 /*     0 */    69,   69,   69,   69,   69,   69,   69,   69,   69,   69,
 /*    10 */    69,   69,   69,   69,   69,   69,   69,   69,   69,   69,
 /*    20 */    69,   69,   69,   69,   69,   93,   86,   85,   88,   87,
 /*    30 */    92,   91,   90,   89,   80,   81,   69,
};
/********** End of lemon-generated parsing tables *****************************/

/* The next table maps tokens (terminal symbols) into fallback tokens.  
** If a construct like the following:
** 
**      %fallback ID X Y Z.
**
** appears in the grammar, then ID becomes a fallback token for X, Y,
** and Z.  Whenever one of the tokens X, Y, or Z is input to the parser
** but it does not parse, the type of the token is changed to ID and
** the parse is retried before an error is thrown.
**
** This feature can be used, for example, to cause some keywords in a language
** to revert to identifiers if they keyword does not apply in the context where
** it appears.
*/
#ifdef YYFALLBACK
static const YYCODETYPE yyFallback[] = {
};
#endif /* YYFALLBACK */

/* The following structure represents a single element of the
** parser's stack.  Information stored includes:
**
**   +  The state number for the parser at this level of the stack.
**
**   +  The value of the token stored at this level of the stack.
**      (In other words, the "major" token.)
**
**   +  The semantic value stored at this level of the stack.  This is
**      the information used by the action routines in the grammar.
**      It is sometimes called the "minor" token.
**
** After the "shift" half of a SHIFTREDUCE action, the stateno field
** actually contains the reduce action for the second half of the
** SHIFTREDUCE.
*/
struct yyStackEntry {
  YYACTIONTYPE stateno;  /* The state-number, or reduce action in SHIFTREDUCE */
  YYCODETYPE major;      /* The major token value.  This is the code
                         ** number for the token at this stack level */
  YYMINORTYPE minor;     /* The user-supplied minor token value.  This
                         ** is the value of the token  */
};
typedef struct yyStackEntry yyStackEntry;

/* The state of the parser is completely contained in an instance of
** the following structure */
struct yyParser {
  yyStackEntry *yytos;          /* Pointer to top element of the stack */
#ifdef YYTRACKMAXSTACKDEPTH
  int yyhwm;                    /* High-water mark of the stack */
#endif
#ifndef YYNOERRORRECOVERY
  int yyerrcnt;                 /* Shifts left before out of the error */
#endif
  ParseARG_SDECL                /* A place to hold %extra_argument */
  ParseCTX_SDECL                /* A place to hold %extra_context */
  yyStackEntry *yystackEnd;           /* Last entry in the stack */
  yyStackEntry *yystack;              /* The parser stack */
  yyStackEntry yystk0[YYSTACKDEPTH];  /* Initial stack space */
};
typedef struct yyParser yyParser;

#include <assert.h>
#ifndef NDEBUG
#include <stdio.h>
static FILE *yyTraceFILE = 0;
static char *yyTracePrompt = 0;
#endif /* NDEBUG */

#ifndef NDEBUG
/* 
** Turn parser tracing on by giving a stream to which to write the trace
** and a prompt to preface each trace message.  Tracing is turned off
** by making either argument NULL 
**
** Inputs:
** <ul>
** <li> A FILE* to which trace output should be written.
**      If NULL, then tracing is turned off.
** <li> A prefix string written at the beginning of every
**      line of trace output.  If NULL, then tracing is
**      turned off.
** </ul>
**
** Outputs:
** None.
*/
void ParseTrace(FILE *TraceFILE, char *zTracePrompt){
  yyTraceFILE = TraceFILE;
  yyTracePrompt = zTracePrompt;
  if( yyTraceFILE==0 ) yyTracePrompt = 0;
  else if( yyTracePrompt==0 ) yyTraceFILE = 0;
}
#endif /* NDEBUG */

#if defined(YYCOVERAGE) || !defined(NDEBUG)
/* For tracing shifts, the names of all terminals and nonterminals
** are required.  The following table supplies these names */
static const char *const yyTokenName[] = { 
  /*    0 */ "$",
  /*    1 */ "NUMBER",
  /*    2 */ "VARIABLE",
  /*    3 */ "LPAREN",
  /*    4 */ "RPAREN",
  /*    5 */ "PLUS",
  /*    6 */ "UPLUS",
  /*    7 */ "MINUS",
  /*    8 */ "UMINUS",
  /*    9 */ "NOT",
  /*   10 */ "FUNCTION_ABS",
  /*   11 */ "MULTIPLY",
  /*   12 */ "DIVIDE",
  /*   13 */ "MODULO",
  /*   14 */ "AND",
  /*   15 */ "OR",
  /*   16 */ "EQ",
  /*   17 */ "NE",
  /*   18 */ "LT",
  /*   19 */ "LE",
  /*   20 */ "GT",
  /*   21 */ "GE",
  /*   22 */ "QMARK",
  /*   23 */ "COLON",
  /*   24 */ "expr",
  /*   25 */ "program",
};
#endif /* defined(YYCOVERAGE) || !defined(NDEBUG) */

#ifndef NDEBUG
/* For tracing reduce actions, the names of all rules are required.
*/
static const char *const yyRuleName[] = {
 /*   0 */ "program ::= expr",
 /*   1 */ "expr ::= NUMBER",
 /*   2 */ "expr ::= VARIABLE",
 /*   3 */ "expr ::= LPAREN expr RPAREN",
 /*   4 */ "expr ::= PLUS expr",
 /*   5 */ "expr ::= MINUS expr",
 /*   6 */ "expr ::= NOT expr",
 /*   7 */ "expr ::= FUNCTION_ABS LPAREN expr RPAREN",
 /*   8 */ "expr ::= expr PLUS expr",
 /*   9 */ "expr ::= expr MINUS expr",
 /*  10 */ "expr ::= expr MULTIPLY expr",
 /*  11 */ "expr ::= expr DIVIDE expr",
 /*  12 */ "expr ::= expr MODULO expr",
 /*  13 */ "expr ::= expr AND expr",
 /*  14 */ "expr ::= expr OR expr",
 /*  15 */ "expr ::= expr EQ expr",
 /*  16 */ "expr ::= expr NE expr",
 /*  17 */ "expr ::= expr LT expr",
 /*  18 */ "expr ::= expr LE expr",
 /*  19 */ "expr ::= expr GT expr",
 /*  20 */ "expr ::= expr GE expr",
 /*  21 */ "expr ::= expr QMARK expr COLON expr",
};
#endif /* NDEBUG */


#if YYGROWABLESTACK
/*
** Try to increase the size of the parser stack.  Return the number
** of errors.  Return 0 on success.
*/
static int yyGrowStack(yyParser *p){
  int oldSize = 1 + (int)(p->yystackEnd - p->yystack);
  int newSize;
  int idx;
  yyStackEntry *pNew;

  newSize = oldSize*2 + 100;
  idx = (int)(p->yytos - p->yystack);
  if( p->yystack==p->yystk0 ){
    pNew = YYREALLOC(0, newSize*sizeof(pNew[0]));
    if( pNew==0 ) return 1;
    memcpy(pNew, p->yystack, oldSize*sizeof(pNew[0]));
  }else{
    pNew = YYREALLOC(p->yystack, newSize*sizeof(pNew[0]));
    if( pNew==0 ) return 1;
  }
  p->yystack = pNew;
  p->yytos = &p->yystack[idx];
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sStack grows from %d to %d entries.\n",
            yyTracePrompt, oldSize, newSize);
  }
#endif
  p->yystackEnd = &p->yystack[newSize-1];
  return 0;
}
#endif /* YYGROWABLESTACK */

#if !YYGROWABLESTACK
/* For builds that do no have a growable stack, yyGrowStack always
** returns an error.
*/
# define yyGrowStack(X) 1
#endif

/* Datatype of the argument to the memory allocated passed as the
** second argument to ParseAlloc() below.  This can be changed by
** putting an appropriate #define in the %include section of the input
** grammar.
*/
#ifndef YYMALLOCARGTYPE
# define YYMALLOCARGTYPE size_t
#endif

/* Initialize a new parser that has already been allocated.
*/
void ParseInit(void *yypRawParser ParseCTX_PDECL){
  yyParser *yypParser = (yyParser*)yypRawParser;
  ParseCTX_STORE
#ifdef YYTRACKMAXSTACKDEPTH
  yypParser->yyhwm = 0;
#endif
  yypParser->yystack = yypParser->yystk0;
  yypParser->yystackEnd = &yypParser->yystack[YYSTACKDEPTH-1];
#ifndef YYNOERRORRECOVERY
  yypParser->yyerrcnt = -1;
#endif
  yypParser->yytos = yypParser->yystack;
  yypParser->yystack[0].stateno = 0;
  yypParser->yystack[0].major = 0;
}

#ifndef Parse_ENGINEALWAYSONSTACK
/* 
** This function allocates a new parser.
** The only argument is a pointer to a function which works like
** malloc.
**
** Inputs:
** A pointer to the function used to allocate memory.
**
** Outputs:
** A pointer to a parser.  This pointer is used in subsequent calls
** to Parse and ParseFree.
*/
void *ParseAlloc(void *(*mallocProc)(YYMALLOCARGTYPE) ParseCTX_PDECL){
  yyParser *yypParser;
  yypParser = (yyParser*)(*mallocProc)( (YYMALLOCARGTYPE)sizeof(yyParser) );
  if( yypParser ){
    ParseCTX_STORE
    ParseInit(yypParser ParseCTX_PARAM);
  }
  return (void*)yypParser;
}
#endif /* Parse_ENGINEALWAYSONSTACK */


/* The following function deletes the "minor type" or semantic value
** associated with a symbol.  The symbol can be either a terminal
** or nonterminal. "yymajor" is the symbol code, and "yypminor" is
** a pointer to the value to be deleted.  The code used to do the 
** deletions is derived from the %destructor and/or %token_destructor
** directives of the input grammar.
*/
static void yy_destructor(
  yyParser *yypParser,    /* The parser */
  YYCODETYPE yymajor,     /* Type code for object to destroy */
  YYMINORTYPE *yypminor   /* The object to be destroyed */
){
  ParseARG_FETCH
  ParseCTX_FETCH
  switch( yymajor ){
    /* Here is inserted the actions which take place when a
    ** terminal or non-terminal is destroyed.  This can happen
    ** when the symbol is popped from the stack during a
    ** reduce or during error processing or when a parser is 
    ** being destroyed before it is finished parsing.
    **
    ** Note: during a reduce, the only symbols destroyed are those
    ** which appear on the RHS of the rule, but which are *not* used
    ** inside the C code.
    */
/********* Begin destructor definitions ***************************************/
    case 24: /* expr */
{
#line 34 "parser.y"

    if ((yypminor->yy48)) {
        eval_node_free((yypminor->yy48));
    }

#line 592 "parser.c"
}
      break;
/********* End destructor definitions *****************************************/
    default:  break;   /* If no destructor action specified: do nothing */
  }
}

/*
** Pop the parser's stack once.
**
** If there is a destructor routine associated with the token which
** is popped from the stack, then call it.
*/
static void yy_pop_parser_stack(yyParser *pParser){
  yyStackEntry *yytos;
  assert( pParser->yytos!=0 );
  assert( pParser->yytos > pParser->yystack );
  yytos = pParser->yytos--;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sPopping %s\n",
      yyTracePrompt,
      yyTokenName[yytos->major]);
  }
#endif
  yy_destructor(pParser, yytos->major, &yytos->minor);
}

/*
** Clear all secondary memory allocations from the parser
*/
void ParseFinalize(void *p){
  yyParser *pParser = (yyParser*)p;

  /* In-lined version of calling yy_pop_parser_stack() for each
  ** element left in the stack */
  yyStackEntry *yytos = pParser->yytos;
  while( yytos>pParser->yystack ){
#ifndef NDEBUG
    if( yyTraceFILE ){
      fprintf(yyTraceFILE,"%sPopping %s\n",
        yyTracePrompt,
        yyTokenName[yytos->major]);
    }
#endif
    if( yytos->major>=YY_MIN_DSTRCTR ){
      yy_destructor(pParser, yytos->major, &yytos->minor);
    }
    yytos--;
  }

#if YYGROWABLESTACK
  if( pParser->yystack!=pParser->yystk0 ) YYFREE(pParser->yystack);
#endif
}

#ifndef Parse_ENGINEALWAYSONSTACK
/* 
** Deallocate and destroy a parser.  Destructors are called for
** all stack elements before shutting the parser down.
**
** If the YYPARSEFREENEVERNULL macro exists (for example because it
** is defined in a %include section of the input grammar) then it is
** assumed that the input pointer is never NULL.
*/
void ParseFree(
  void *p,                    /* The parser to be deleted */
  void (*freeProc)(void*)     /* Function used to reclaim memory */
){
#ifndef YYPARSEFREENEVERNULL
  if( p==0 ) return;
#endif
  ParseFinalize(p);
  (*freeProc)(p);
}
#endif /* Parse_ENGINEALWAYSONSTACK */

/*
** Return the peak depth of the stack for a parser.
*/
#ifdef YYTRACKMAXSTACKDEPTH
int ParseStackPeak(void *p){
  yyParser *pParser = (yyParser*)p;
  return pParser->yyhwm;
}
#endif

/* This array of booleans keeps track of the parser statement
** coverage.  The element yycoverage[X][Y] is set when the parser
** is in state X and has a lookahead token Y.  In a well-tested
** systems, every element of this matrix should end up being set.
*/
#if defined(YYCOVERAGE)
static unsigned char yycoverage[YYNSTATE][YYNTOKEN];
#endif

/*
** Write into out a description of every state/lookahead combination that
**
**   (1)  has not been used by the parser, and
**   (2)  is not a syntax error.
**
** Return the number of missed state/lookahead combinations.
*/
#if defined(YYCOVERAGE)
int ParseCoverage(FILE *out){
  int stateno, iLookAhead, i;
  int nMissed = 0;
  for(stateno=0; stateno<YYNSTATE; stateno++){
    i = yy_shift_ofst[stateno];
    for(iLookAhead=0; iLookAhead<YYNTOKEN; iLookAhead++){
      if( yy_lookahead[i+iLookAhead]!=iLookAhead ) continue;
      if( yycoverage[stateno][iLookAhead]==0 ) nMissed++;
      if( out ){
        fprintf(out,"State %d lookahead %s %s\n", stateno,
                yyTokenName[iLookAhead],
                yycoverage[stateno][iLookAhead] ? "ok" : "missed");
      }
    }
  }
  return nMissed;
}
#endif

/*
** Find the appropriate action for a parser given the terminal
** look-ahead token iLookAhead.
*/
static YYACTIONTYPE yy_find_shift_action(
  YYCODETYPE iLookAhead,    /* The look-ahead token */
  YYACTIONTYPE stateno      /* Current state number */
){
  int i;

  if( stateno>YY_MAX_SHIFT ) return stateno;
  assert( stateno <= YY_SHIFT_COUNT );
#if defined(YYCOVERAGE)
  yycoverage[stateno][iLookAhead] = 1;
#endif
  do{
    i = yy_shift_ofst[stateno];
    assert( i>=0 );
    assert( i<=YY_ACTTAB_COUNT );
    assert( i+YYNTOKEN<=(int)YY_NLOOKAHEAD );
    assert( iLookAhead!=YYNOCODE );
    assert( iLookAhead < YYNTOKEN );
    i += iLookAhead;
    assert( i<(int)YY_NLOOKAHEAD );
    if( yy_lookahead[i]!=iLookAhead ){
#ifdef YYFALLBACK
      YYCODETYPE iFallback;            /* Fallback token */
      assert( iLookAhead<sizeof(yyFallback)/sizeof(yyFallback[0]) );
      iFallback = yyFallback[iLookAhead];
      if( iFallback!=0 ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE, "%sFALLBACK %s => %s\n",
             yyTracePrompt, yyTokenName[iLookAhead], yyTokenName[iFallback]);
        }
#endif
        assert( yyFallback[iFallback]==0 ); /* Fallback loop must terminate */
        iLookAhead = iFallback;
        continue;
      }
#endif
#ifdef YYWILDCARD
      {
        int j = i - iLookAhead + YYWILDCARD;
        assert( j<(int)(sizeof(yy_lookahead)/sizeof(yy_lookahead[0])) );
        if( yy_lookahead[j]==YYWILDCARD && iLookAhead>0 ){
#ifndef NDEBUG
          if( yyTraceFILE ){
            fprintf(yyTraceFILE, "%sWILDCARD %s => %s\n",
               yyTracePrompt, yyTokenName[iLookAhead],
               yyTokenName[YYWILDCARD]);
          }
#endif /* NDEBUG */
          return yy_action[j];
        }
      }
#endif /* YYWILDCARD */
      return yy_default[stateno];
    }else{
      assert( i>=0 && i<(int)(sizeof(yy_action)/sizeof(yy_action[0])) );
      return yy_action[i];
    }
  }while(1);
}

/*
** Find the appropriate action for a parser given the non-terminal
** look-ahead token iLookAhead.
*/
static YYACTIONTYPE yy_find_reduce_action(
  YYACTIONTYPE stateno,     /* Current state number */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
#ifdef YYERRORSYMBOL
  if( stateno>YY_REDUCE_COUNT ){
    return yy_default[stateno];
  }
#else
  assert( stateno<=YY_REDUCE_COUNT );
#endif
  i = yy_reduce_ofst[stateno];
  assert( iLookAhead!=YYNOCODE );
  i += iLookAhead;
#ifdef YYERRORSYMBOL
  if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
    return yy_default[stateno];
  }
#else
  assert( i>=0 && i<YY_ACTTAB_COUNT );
  assert( yy_lookahead[i]==iLookAhead );
#endif
  return yy_action[i];
}

/*
** The following routine is called if the stack overflows.
*/
static void yyStackOverflow(yyParser *yypParser){
   ParseARG_FETCH
   ParseCTX_FETCH
#ifndef NDEBUG
   if( yyTraceFILE ){
     fprintf(yyTraceFILE,"%sStack Overflow!\n",yyTracePrompt);
   }
#endif
   while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
   /* Here code is inserted which will execute if the parser
   ** stack every overflows */
/******** Begin %stack_overflow code ******************************************/
/******** End %stack_overflow code ********************************************/
   ParseARG_STORE /* Suppress warning about unused %extra_argument var */
   ParseCTX_STORE
}

/*
** Print tracing information for a SHIFT action
*/
#ifndef NDEBUG
static void yyTraceShift(yyParser *yypParser, int yyNewState, const char *zTag){
  if( yyTraceFILE ){
    if( yyNewState<YYNSTATE ){
      fprintf(yyTraceFILE,"%s%s '%s', go to state %d\n",
         yyTracePrompt, zTag, yyTokenName[yypParser->yytos->major],
         yyNewState);
    }else{
      fprintf(yyTraceFILE,"%s%s '%s', pending reduce %d\n",
         yyTracePrompt, zTag, yyTokenName[yypParser->yytos->major],
         yyNewState - YY_MIN_REDUCE);
    }
  }
}
#else
# define yyTraceShift(X,Y,Z)
#endif

/*
** Perform a shift action.
*/
static void yy_shift(
  yyParser *yypParser,          /* The parser to be shifted */
  YYACTIONTYPE yyNewState,      /* The new state to shift in */
  YYCODETYPE yyMajor,           /* The major token to shift in */
  ParseTOKENTYPE yyMinor        /* The minor token to shift in */
){
  yyStackEntry *yytos;
  yypParser->yytos++;
#ifdef YYTRACKMAXSTACKDEPTH
  if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
    yypParser->yyhwm++;
    assert( yypParser->yyhwm == (int)(yypParser->yytos - yypParser->yystack) );
  }
#endif
  yytos = yypParser->yytos;
  if( yytos>yypParser->yystackEnd ){
    if( yyGrowStack(yypParser) ){
      yypParser->yytos--;
      yyStackOverflow(yypParser);
      return;
    }
    yytos = yypParser->yytos;
    assert( yytos <= yypParser->yystackEnd );
  }
  if( yyNewState > YY_MAX_SHIFT ){
    yyNewState += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
  }
  yytos->stateno = yyNewState;
  yytos->major = yyMajor;
  yytos->minor.yy0 = yyMinor;
  yyTraceShift(yypParser, yyNewState, "Shift");
}

/* For rule J, yyRuleInfoLhs[J] contains the symbol on the left-hand side
** of that rule */
static const YYCODETYPE yyRuleInfoLhs[] = {
    25,  /* (0) program ::= expr */
    24,  /* (1) expr ::= NUMBER */
    24,  /* (2) expr ::= VARIABLE */
    24,  /* (3) expr ::= LPAREN expr RPAREN */
    24,  /* (4) expr ::= PLUS expr */
    24,  /* (5) expr ::= MINUS expr */
    24,  /* (6) expr ::= NOT expr */
    24,  /* (7) expr ::= FUNCTION_ABS LPAREN expr RPAREN */
    24,  /* (8) expr ::= expr PLUS expr */
    24,  /* (9) expr ::= expr MINUS expr */
    24,  /* (10) expr ::= expr MULTIPLY expr */
    24,  /* (11) expr ::= expr DIVIDE expr */
    24,  /* (12) expr ::= expr MODULO expr */
    24,  /* (13) expr ::= expr AND expr */
    24,  /* (14) expr ::= expr OR expr */
    24,  /* (15) expr ::= expr EQ expr */
    24,  /* (16) expr ::= expr NE expr */
    24,  /* (17) expr ::= expr LT expr */
    24,  /* (18) expr ::= expr LE expr */
    24,  /* (19) expr ::= expr GT expr */
    24,  /* (20) expr ::= expr GE expr */
    24,  /* (21) expr ::= expr QMARK expr COLON expr */
};

/* For rule J, yyRuleInfoNRhs[J] contains the negative of the number
** of symbols on the right-hand side of that rule. */
static const signed char yyRuleInfoNRhs[] = {
   -1,  /* (0) program ::= expr */
   -1,  /* (1) expr ::= NUMBER */
   -1,  /* (2) expr ::= VARIABLE */
   -3,  /* (3) expr ::= LPAREN expr RPAREN */
   -2,  /* (4) expr ::= PLUS expr */
   -2,  /* (5) expr ::= MINUS expr */
   -2,  /* (6) expr ::= NOT expr */
   -4,  /* (7) expr ::= FUNCTION_ABS LPAREN expr RPAREN */
   -3,  /* (8) expr ::= expr PLUS expr */
   -3,  /* (9) expr ::= expr MINUS expr */
   -3,  /* (10) expr ::= expr MULTIPLY expr */
   -3,  /* (11) expr ::= expr DIVIDE expr */
   -3,  /* (12) expr ::= expr MODULO expr */
   -3,  /* (13) expr ::= expr AND expr */
   -3,  /* (14) expr ::= expr OR expr */
   -3,  /* (15) expr ::= expr EQ expr */
   -3,  /* (16) expr ::= expr NE expr */
   -3,  /* (17) expr ::= expr LT expr */
   -3,  /* (18) expr ::= expr LE expr */
   -3,  /* (19) expr ::= expr GT expr */
   -3,  /* (20) expr ::= expr GE expr */
   -5,  /* (21) expr ::= expr QMARK expr COLON expr */
};

static void yy_accept(yyParser*);  /* Forward Declaration */

/*
** Perform a reduce action and the shift that must immediately
** follow the reduce.
**
** The yyLookahead and yyLookaheadToken parameters provide reduce actions
** access to the lookahead token (if any).  The yyLookahead will be YYNOCODE
** if the lookahead token has already been consumed.  As this procedure is
** only called from one place, optimizing compilers will in-line it, which
** means that the extra parameters have no performance impact.
*/
static YYACTIONTYPE yy_reduce(
  yyParser *yypParser,         /* The parser */
  unsigned int yyruleno,       /* Number of the rule by which to reduce */
  int yyLookahead,             /* Lookahead token, or YYNOCODE if none */
  ParseTOKENTYPE yyLookaheadToken  /* Value of the lookahead token */
  ParseCTX_PDECL                   /* %extra_context */
){
  int yygoto;                     /* The next state */
  YYACTIONTYPE yyact;             /* The next action */
  yyStackEntry *yymsp;            /* The top of the parser's stack */
  int yysize;                     /* Amount to pop the stack */
  ParseARG_FETCH
  (void)yyLookahead;
  (void)yyLookaheadToken;
  yymsp = yypParser->yytos;

  switch( yyruleno ){
  /* Beginning here are the reduction cases.  A typical example
  ** follows:
  **   case 0:
  **  #line <lineno> <grammarfile>
  **     { ... }           // User supplied code
  **  #line <lineno> <thisfile>
  **     break;
  */
/********** Begin reduce actions **********************************************/
        YYMINORTYPE yylhsminor;
      case 0: /* program ::= expr */
#line 41 "parser.y"
{
    *result = yymsp[0].minor.yy48;
}
#line 987 "parser.c"
        break;
      case 1: /* expr ::= NUMBER */
#line 46 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(1);
    yylhsminor.yy48->operator = EVAL_OPERATOR_NOP;
    eval_node_set_value_to_constant(yylhsminor.yy48, 0, yymsp[0].minor.yy0.dval);
}
#line 996 "parser.c"
  yymsp[0].minor.yy48 = yylhsminor.yy48;
        break;
      case 2: /* expr ::= VARIABLE */
#line 52 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(1);
    yylhsminor.yy48->operator = EVAL_OPERATOR_NOP;
    eval_node_set_value_to_variable(yylhsminor.yy48, 0, yymsp[0].minor.yy0.strval);
    freez(yymsp[0].minor.yy0.strval); // Free the strdup'd string
}
#line 1007 "parser.c"
  yymsp[0].minor.yy48 = yylhsminor.yy48;
        break;
      case 3: /* expr ::= LPAREN expr RPAREN */
#line 60 "parser.y"
{
    yymsp[-2].minor.yy48 = eval_node_alloc(1);
    yymsp[-2].minor.yy48->operator = EVAL_OPERATOR_EXPRESSION_OPEN;
    yymsp[-2].minor.yy48->precedence = eval_precedence(EVAL_OPERATOR_EXPRESSION_OPEN);
    eval_node_set_value_to_node(yymsp[-2].minor.yy48, 0, yymsp[-1].minor.yy48);
}
#line 1018 "parser.c"
        break;
      case 4: /* expr ::= PLUS expr */
#line 68 "parser.y"
{
    yymsp[-1].minor.yy48 = eval_node_alloc(1);
    yymsp[-1].minor.yy48->operator = EVAL_OPERATOR_SIGN_PLUS;
    yymsp[-1].minor.yy48->precedence = eval_precedence(EVAL_OPERATOR_SIGN_PLUS);
    eval_node_set_value_to_node(yymsp[-1].minor.yy48, 0, yymsp[0].minor.yy48);
}
#line 1028 "parser.c"
        break;
      case 5: /* expr ::= MINUS expr */
#line 75 "parser.y"
{
    yymsp[-1].minor.yy48 = eval_node_alloc(1);
    yymsp[-1].minor.yy48->operator = EVAL_OPERATOR_SIGN_MINUS;
    yymsp[-1].minor.yy48->precedence = eval_precedence(EVAL_OPERATOR_SIGN_MINUS);
    eval_node_set_value_to_node(yymsp[-1].minor.yy48, 0, yymsp[0].minor.yy48);
}
#line 1038 "parser.c"
        break;
      case 6: /* expr ::= NOT expr */
#line 82 "parser.y"
{
    yymsp[-1].minor.yy48 = eval_node_alloc(1);
    yymsp[-1].minor.yy48->operator = EVAL_OPERATOR_NOT;
    yymsp[-1].minor.yy48->precedence = eval_precedence(EVAL_OPERATOR_NOT);
    eval_node_set_value_to_node(yymsp[-1].minor.yy48, 0, yymsp[0].minor.yy48);
}
#line 1048 "parser.c"
        break;
      case 7: /* expr ::= FUNCTION_ABS LPAREN expr RPAREN */
#line 90 "parser.y"
{
    yymsp[-3].minor.yy48 = eval_node_alloc(1);
    yymsp[-3].minor.yy48->operator = EVAL_OPERATOR_ABS;
    yymsp[-3].minor.yy48->precedence = eval_precedence(EVAL_OPERATOR_ABS);
    eval_node_set_value_to_node(yymsp[-3].minor.yy48, 0, yymsp[-1].minor.yy48);
}
#line 1058 "parser.c"
        break;
      case 8: /* expr ::= expr PLUS expr */
#line 98 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_PLUS;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_PLUS);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1069 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 9: /* expr ::= expr MINUS expr */
#line 106 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_MINUS;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_MINUS);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1081 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 10: /* expr ::= expr MULTIPLY expr */
#line 114 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_MULTIPLY;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_MULTIPLY);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1093 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 11: /* expr ::= expr DIVIDE expr */
#line 122 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_DIVIDE;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_DIVIDE);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1105 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 12: /* expr ::= expr MODULO expr */
#line 130 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_MODULO;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_MODULO);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1117 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 13: /* expr ::= expr AND expr */
#line 138 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_AND;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_AND);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1129 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 14: /* expr ::= expr OR expr */
#line 146 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_OR;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_OR);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1141 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 15: /* expr ::= expr EQ expr */
#line 154 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_EQUAL;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_EQUAL);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1153 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 16: /* expr ::= expr NE expr */
#line 162 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_NOT_EQUAL;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_NOT_EQUAL);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1165 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 17: /* expr ::= expr LT expr */
#line 170 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_LESS;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_LESS);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1177 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 18: /* expr ::= expr LE expr */
#line 178 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_LESS_THAN_OR_EQUAL;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_LESS_THAN_OR_EQUAL);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1189 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 19: /* expr ::= expr GT expr */
#line 186 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_GREATER;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_GREATER);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1201 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 20: /* expr ::= expr GE expr */
#line 194 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(2);
    yylhsminor.yy48->operator = EVAL_OPERATOR_GREATER_THAN_OR_EQUAL;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_GREATER_THAN_OR_EQUAL);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[0].minor.yy48);
}
#line 1213 "parser.c"
  yymsp[-2].minor.yy48 = yylhsminor.yy48;
        break;
      case 21: /* expr ::= expr QMARK expr COLON expr */
#line 205 "parser.y"
{
    yylhsminor.yy48 = eval_node_alloc(3);
    yylhsminor.yy48->operator = EVAL_OPERATOR_IF_THEN_ELSE;
    yylhsminor.yy48->precedence = eval_precedence(EVAL_OPERATOR_IF_THEN_ELSE);
    eval_node_set_value_to_node(yylhsminor.yy48, 0, yymsp[-4].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 1, yymsp[-2].minor.yy48);
    eval_node_set_value_to_node(yylhsminor.yy48, 2, yymsp[0].minor.yy48);
}
#line 1226 "parser.c"
  yymsp[-4].minor.yy48 = yylhsminor.yy48;
        break;
      default:
        break;
/********** End reduce actions ************************************************/
  };
  assert( yyruleno<sizeof(yyRuleInfoLhs)/sizeof(yyRuleInfoLhs[0]) );
  yygoto = yyRuleInfoLhs[yyruleno];
  yysize = yyRuleInfoNRhs[yyruleno];
  yyact = yy_find_reduce_action(yymsp[yysize].stateno,(YYCODETYPE)yygoto);

  /* There are no SHIFTREDUCE actions on nonterminals because the table
  ** generator has simplified them to pure REDUCE actions. */
  assert( !(yyact>YY_MAX_SHIFT && yyact<=YY_MAX_SHIFTREDUCE) );

  /* It is not possible for a REDUCE to be followed by an error */
  assert( yyact!=YY_ERROR_ACTION );

  yymsp += yysize+1;
  yypParser->yytos = yymsp;
  yymsp->stateno = (YYACTIONTYPE)yyact;
  yymsp->major = (YYCODETYPE)yygoto;
  yyTraceShift(yypParser, yyact, "... then shift");
  return yyact;
}

/*
** The following code executes when the parse fails
*/
#ifndef YYNOERRORRECOVERY
static void yy_parse_failed(
  yyParser *yypParser           /* The parser */
){
  ParseARG_FETCH
  ParseCTX_FETCH
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sFail!\n",yyTracePrompt);
  }
#endif
  while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
  /* Here code is inserted which will be executed whenever the
  ** parser fails */
/************ Begin %parse_failure code ***************************************/
#line 24 "parser.y"

    // Failed to parse the expression
    if (*result) {
        eval_node_free(*result);
        *result = NULL;
    }
#line 1278 "parser.c"
/************ End %parse_failure code *****************************************/
  ParseARG_STORE /* Suppress warning about unused %extra_argument variable */
  ParseCTX_STORE
}
#endif /* YYNOERRORRECOVERY */

/*
** The following code executes when a syntax error first occurs.
*/
static void yy_syntax_error(
  yyParser *yypParser,           /* The parser */
  int yymajor,                   /* The major type of the error token */
  ParseTOKENTYPE yyminor         /* The minor type of the error token */
){
  ParseARG_FETCH
  ParseCTX_FETCH
#define TOKEN yyminor
/************ Begin %syntax_error code ****************************************/
#line 13 "parser.y"

    // Create a NOP node with count=0 as an error marker
    EVAL_NODE *error_node = eval_node_alloc(0);
    error_node->operator = EVAL_OPERATOR_NOP;
    *result = error_node;
#line 1303 "parser.c"
/************ End %syntax_error code ******************************************/
  ParseARG_STORE /* Suppress warning about unused %extra_argument variable */
  ParseCTX_STORE
}

/*
** The following is executed when the parser accepts
*/
static void yy_accept(
  yyParser *yypParser           /* The parser */
){
  ParseARG_FETCH
  ParseCTX_FETCH
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sAccept!\n",yyTracePrompt);
  }
#endif
#ifndef YYNOERRORRECOVERY
  yypParser->yyerrcnt = -1;
#endif
  assert( yypParser->yytos==yypParser->yystack );
  /* Here code is inserted which will be executed whenever the
  ** parser accepts */
/*********** Begin %parse_accept code *****************************************/
#line 20 "parser.y"

    // Successfully parsed the expression
#line 1332 "parser.c"
/*********** End %parse_accept code *******************************************/
  ParseARG_STORE /* Suppress warning about unused %extra_argument variable */
  ParseCTX_STORE
}

/* The main parser program.
** The first argument is a pointer to a structure obtained from
** "ParseAlloc" which describes the current state of the parser.
** The second argument is the major token number.  The third is
** the minor token.  The fourth optional argument is whatever the
** user wants (and specified in the grammar) and is available for
** use by the action routines.
**
** Inputs:
** <ul>
** <li> A pointer to the parser (an opaque structure.)
** <li> The major token number.
** <li> The minor token number.
** <li> An option argument of a grammar-specified type.
** </ul>
**
** Outputs:
** None.
*/
void Parse(
  void *yyp,                   /* The parser */
  int yymajor,                 /* The major token code number */
  ParseTOKENTYPE yyminor       /* The value for the token */
  ParseARG_PDECL               /* Optional %extra_argument parameter */
){
  YYMINORTYPE yyminorunion;
  YYACTIONTYPE yyact;   /* The parser action. */
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  int yyendofinput;     /* True if we are at the end of input */
#endif
#ifdef YYERRORSYMBOL
  int yyerrorhit = 0;   /* True if yymajor has invoked an error */
#endif
  yyParser *yypParser = (yyParser*)yyp;  /* The parser */
  ParseCTX_FETCH
  ParseARG_STORE

  assert( yypParser->yytos!=0 );
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  yyendofinput = (yymajor==0);
#endif

  yyact = yypParser->yytos->stateno;
#ifndef NDEBUG
  if( yyTraceFILE ){
    if( yyact < YY_MIN_REDUCE ){
      fprintf(yyTraceFILE,"%sInput '%s' in state %d\n",
              yyTracePrompt,yyTokenName[yymajor],yyact);
    }else{
      fprintf(yyTraceFILE,"%sInput '%s' with pending reduce %d\n",
              yyTracePrompt,yyTokenName[yymajor],yyact-YY_MIN_REDUCE);
    }
  }
#endif

  while(1){ /* Exit by "break" */
    assert( yypParser->yytos>=yypParser->yystack );
    assert( yyact==yypParser->yytos->stateno );
    yyact = yy_find_shift_action((YYCODETYPE)yymajor,yyact);
    if( yyact >= YY_MIN_REDUCE ){
      unsigned int yyruleno = yyact - YY_MIN_REDUCE; /* Reduce by this rule */
#ifndef NDEBUG
      assert( yyruleno<(int)(sizeof(yyRuleName)/sizeof(yyRuleName[0])) );
      if( yyTraceFILE ){
        int yysize = yyRuleInfoNRhs[yyruleno];
        if( yysize ){
          fprintf(yyTraceFILE, "%sReduce %d [%s]%s, pop back to state %d.\n",
            yyTracePrompt,
            yyruleno, yyRuleName[yyruleno],
            yyruleno<YYNRULE_WITH_ACTION ? "" : " without external action",
            yypParser->yytos[yysize].stateno);
        }else{
          fprintf(yyTraceFILE, "%sReduce %d [%s]%s.\n",
            yyTracePrompt, yyruleno, yyRuleName[yyruleno],
            yyruleno<YYNRULE_WITH_ACTION ? "" : " without external action");
        }
      }
#endif /* NDEBUG */

      /* Check that the stack is large enough to grow by a single entry
      ** if the RHS of the rule is empty.  This ensures that there is room
      ** enough on the stack to push the LHS value */
      if( yyRuleInfoNRhs[yyruleno]==0 ){
#ifdef YYTRACKMAXSTACKDEPTH
        if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
          yypParser->yyhwm++;
          assert( yypParser->yyhwm ==
                  (int)(yypParser->yytos - yypParser->yystack));
        }
#endif
        if( yypParser->yytos>=yypParser->yystackEnd ){
          if( yyGrowStack(yypParser) ){
            yyStackOverflow(yypParser);
            break;
          }
        }
      }
      yyact = yy_reduce(yypParser,yyruleno,yymajor,yyminor ParseCTX_PARAM);
    }else if( yyact <= YY_MAX_SHIFTREDUCE ){
      yy_shift(yypParser,yyact,(YYCODETYPE)yymajor,yyminor);
#ifndef YYNOERRORRECOVERY
      yypParser->yyerrcnt--;
#endif
      break;
    }else if( yyact==YY_ACCEPT_ACTION ){
      yypParser->yytos--;
      yy_accept(yypParser);
      return;
    }else{
      assert( yyact == YY_ERROR_ACTION );
      yyminorunion.yy0 = yyminor;
#ifdef YYERRORSYMBOL
      int yymx;
#endif
#ifndef NDEBUG
      if( yyTraceFILE ){
        fprintf(yyTraceFILE,"%sSyntax Error!\n",yyTracePrompt);
      }
#endif
#ifdef YYERRORSYMBOL
      /* A syntax error has occurred.
      ** The response to an error depends upon whether or not the
      ** grammar defines an error token "ERROR".  
      **
      ** This is what we do if the grammar does define ERROR:
      **
      **  * Call the %syntax_error function.
      **
      **  * Begin popping the stack until we enter a state where
      **    it is legal to shift the error symbol, then shift
      **    the error symbol.
      **
      **  * Set the error count to three.
      **
      **  * Begin accepting and shifting new tokens.  No new error
      **    processing will occur until three tokens have been
      **    shifted successfully.
      **
      */
      if( yypParser->yyerrcnt<0 ){
        yy_syntax_error(yypParser,yymajor,yyminor);
      }
      yymx = yypParser->yytos->major;
      if( yymx==YYERRORSYMBOL || yyerrorhit ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE,"%sDiscard input token %s\n",
             yyTracePrompt,yyTokenName[yymajor]);
        }
#endif
        yy_destructor(yypParser, (YYCODETYPE)yymajor, &yyminorunion);
        yymajor = YYNOCODE;
      }else{
        while( yypParser->yytos > yypParser->yystack ){
          yyact = yy_find_reduce_action(yypParser->yytos->stateno,
                                        YYERRORSYMBOL);
          if( yyact<=YY_MAX_SHIFTREDUCE ) break;
          yy_pop_parser_stack(yypParser);
        }
        if( yypParser->yytos <= yypParser->yystack || yymajor==0 ){
          yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
          yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
          yypParser->yyerrcnt = -1;
#endif
          yymajor = YYNOCODE;
        }else if( yymx!=YYERRORSYMBOL ){
          yy_shift(yypParser,yyact,YYERRORSYMBOL,yyminor);
        }
      }
      yypParser->yyerrcnt = 3;
      yyerrorhit = 1;
      if( yymajor==YYNOCODE ) break;
      yyact = yypParser->yytos->stateno;
#elif defined(YYNOERRORRECOVERY)
      /* If the YYNOERRORRECOVERY macro is defined, then do not attempt to
      ** do any kind of error recovery.  Instead, simply invoke the syntax
      ** error routine and continue going as if nothing had happened.
      **
      ** Applications can set this macro (for example inside %include) if
      ** they intend to abandon the parse upon the first syntax error seen.
      */
      yy_syntax_error(yypParser,yymajor, yyminor);
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      break;
#else  /* YYERRORSYMBOL is not defined */
      /* This is what we do if the grammar does not define ERROR:
      **
      **  * Report an error message, and throw away the input token.
      **
      **  * If the input token is $, then fail the parse.
      **
      ** As before, subsequent error messages are suppressed until
      ** three input tokens have been successfully shifted.
      */
      if( yypParser->yyerrcnt<=0 ){
        yy_syntax_error(yypParser,yymajor, yyminor);
      }
      yypParser->yyerrcnt = 3;
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      if( yyendofinput ){
        yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
        yypParser->yyerrcnt = -1;
#endif
      }
      break;
#endif
    }
  }
#ifndef NDEBUG
  if( yyTraceFILE ){
    yyStackEntry *i;
    char cDiv = '[';
    fprintf(yyTraceFILE,"%sReturn. Stack=",yyTracePrompt);
    for(i=&yypParser->yystack[1]; i<=yypParser->yytos; i++){
      fprintf(yyTraceFILE,"%c%s", cDiv, yyTokenName[i->major]);
      cDiv = ' ';
    }
    fprintf(yyTraceFILE,"]\n");
  }
#endif
  return;
}

/*
** Return the fallback token corresponding to canonical token iToken, or
** 0 if iToken has no fallback.
*/
int ParseFallback(int iToken){
#ifdef YYFALLBACK
  assert( iToken<(int)(sizeof(yyFallback)/sizeof(yyFallback[0])) );
  return yyFallback[iToken];
#else
  (void)iToken;
  return 0;
#endif
}
