// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: proto/agent/v1/cmds.proto
// Protobuf C++ Version: 7.35.1

#ifndef proto_2fagent_2fv1_2fcmds_2eproto_2epb_2eh
#define proto_2fagent_2fv1_2fcmds_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/unknown_field_set.h"
#include "google/protobuf/timestamp.pb.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_proto_2fagent_2fv1_2fcmds_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_proto_2fagent_2fv1_2fcmds_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_proto_2fagent_2fv1_2fcmds_2eproto;
}  // extern "C"
namespace agent {
namespace v1 {
class AgentCommand;
struct AgentCommandGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern AgentCommandGlobalsTypeInternal AgentCommand_globals_;
extern const ::google::protobuf::internal::ClassDataFull AgentCommand_class_data_;
#else
extern const AgentCommandGlobalsTypeInternal AgentCommand_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class AgentCommandAck;
struct AgentCommandAckGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern AgentCommandAckGlobalsTypeInternal AgentCommandAck_globals_;
extern const ::google::protobuf::internal::ClassDataFull AgentCommandAck_class_data_;
#else
extern const AgentCommandAckGlobalsTypeInternal AgentCommandAck_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class AgentCommandResponse;
struct AgentCommandResponseGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern AgentCommandResponseGlobalsTypeInternal AgentCommandResponse_globals_;
extern const ::google::protobuf::internal::ClassDataFull AgentCommandResponse_class_data_;
#else
extern const AgentCommandResponseGlobalsTypeInternal AgentCommandResponse_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class CancelPendingRequest;
struct CancelPendingRequestGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern CancelPendingRequestGlobalsTypeInternal CancelPendingRequest_globals_;
extern const ::google::protobuf::internal::ClassDataFull CancelPendingRequest_class_data_;
#else
extern const CancelPendingRequestGlobalsTypeInternal CancelPendingRequest_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace v1
}  // namespace agent
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google

namespace agent {
namespace v1 {

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED AgentCommand final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:agent.v1.AgentCommand) */ {
 public:
  inline AgentCommand() : AgentCommand(nullptr) {}
  ~AgentCommand() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(AgentCommand* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(AgentCommand));
  }
#endif

  template <typename = void>
  explicit constexpr AgentCommand(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline AgentCommand(const AgentCommand& from) : AgentCommand(nullptr, from) {}
  inline AgentCommand(AgentCommand&& from) noexcept : AgentCommand(nullptr, ::std::move(from)) {}
  inline AgentCommand& operator=(const AgentCommand& from) {
    CopyFrom(from);
    return *this;
  }
  inline AgentCommand& operator=(AgentCommand&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const AgentCommand& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<AgentCommand>(&AgentCommand_globals_);
  }
  static constexpr int kIndexInFileMessages = 1;
  friend void swap(AgentCommand& a, AgentCommand& b) { a.Swap(&b); }
  inline void Swap(AgentCommand* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(AgentCommand* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] AgentCommand* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<AgentCommand>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const AgentCommand& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const AgentCommand& from) { AgentCommand::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(AgentCommand* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "agent.v1.AgentCommand"; }

  explicit AgentCommand(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  AgentCommand(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const AgentCommand& from);
  AgentCommand(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, AgentCommand&& from) noexcept
      : AgentCommand(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kCallbackTopicFieldNumber = 1,
    kAckTopicFieldNumber = 2,
    kMessageIdFieldNumber = 3,
    kResourceFieldNumber = 6,
    kTimeoutFieldNumber = 4,
    kAckTimeoutFieldNumber = 5,
  };
  // string callback_topic = 1;
  void clear_callback_topic() ;
  [[nodiscard]] const ::std::string& callback_topic() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_callback_topic(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_callback_topic();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_callback_topic();
  void set_allocated_callback_topic(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_callback_topic() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_callback_topic(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_callback_topic();

  public:
  // string ack_topic = 2;
  void clear_ack_topic() ;
  [[nodiscard]] const ::std::string& ack_topic() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_ack_topic(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_ack_topic();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_ack_topic();
  void set_allocated_ack_topic(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_ack_topic() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_ack_topic(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_ack_topic();

  public:
  // string message_id = 3;
  void clear_message_id() ;
  [[nodiscard]] const ::std::string& message_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_message_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_message_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_message_id();
  void set_allocated_message_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_message_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_message_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_message_id();

  public:
  // string resource = 6;
  void clear_resource() ;
  [[nodiscard]] const ::std::string& resource() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_resource(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_resource();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_resource();
  void set_allocated_resource(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_resource() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_resource(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_resource();

  public:
  // uint64 timeout = 4;
  void clear_timeout() ;
  [[nodiscard]] ::uint64_t timeout() const;
  void set_timeout(::uint64_t value);

  private:
  ::uint64_t _internal_timeout() const;
  void _internal_set_timeout(::uint64_t value);

  public:
  // uint64 ack_timeout = 5;
  void clear_ack_timeout() ;
  [[nodiscard]] ::uint64_t ack_timeout() const;
  void set_ack_timeout(::uint64_t value);

  private:
  ::uint64_t _internal_ack_timeout() const;
  void _internal_set_ack_timeout(::uint64_t value);

  public:
  // @@protoc_insertion_point(class_scope:agent.v1.AgentCommand)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<3, 6,
                          0, 71,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const AgentCommand& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr callback_topic_;
    ::google::protobuf::internal::ArenaStringPtr ack_topic_;
    ::google::protobuf::internal::ArenaStringPtr message_id_;
    ::google::protobuf::internal::ArenaStringPtr resource_;
    ::uint64_t timeout_;
    ::uint64_t ack_timeout_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_proto_2fagent_2fv1_2fcmds_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED CancelPendingRequest final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:agent.v1.CancelPendingRequest) */ {
 public:
  inline CancelPendingRequest() : CancelPendingRequest(nullptr) {}
  ~CancelPendingRequest() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(CancelPendingRequest* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(CancelPendingRequest));
  }
#endif

  template <typename = void>
  explicit constexpr CancelPendingRequest(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline CancelPendingRequest(const CancelPendingRequest& from) : CancelPendingRequest(nullptr, from) {}
  inline CancelPendingRequest(CancelPendingRequest&& from) noexcept : CancelPendingRequest(nullptr, ::std::move(from)) {}
  inline CancelPendingRequest& operator=(const CancelPendingRequest& from) {
    CopyFrom(from);
    return *this;
  }
  inline CancelPendingRequest& operator=(CancelPendingRequest&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const CancelPendingRequest& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<CancelPendingRequest>(&CancelPendingRequest_globals_);
  }
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(CancelPendingRequest& a, CancelPendingRequest& b) { a.Swap(&b); }
  inline void Swap(CancelPendingRequest* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(CancelPendingRequest* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] CancelPendingRequest* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<CancelPendingRequest>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const CancelPendingRequest& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const CancelPendingRequest& from) { CancelPendingRequest::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(CancelPendingRequest* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "agent.v1.CancelPendingRequest"; }

  explicit CancelPendingRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  CancelPendingRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const CancelPendingRequest& from);
  CancelPendingRequest(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, CancelPendingRequest&& from) noexcept
      : CancelPendingRequest(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kRequestIdFieldNumber = 1,
    kTraceIdFieldNumber = 3,
    kTimestampFieldNumber = 2,
  };
  // string request_id = 1;
  void clear_request_id() ;
  [[nodiscard]] const ::std::string& request_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_request_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_request_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_request_id();
  void set_allocated_request_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_request_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_request_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_request_id();

  public:
  // string trace_id = 3;
  void clear_trace_id() ;
  [[nodiscard]] const ::std::string& trace_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_trace_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_trace_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_trace_id();
  void set_allocated_trace_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_trace_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_trace_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_trace_id();

  public:
  // .google.protobuf.Timestamp timestamp = 2;
  [[nodiscard]] bool has_timestamp()
      const;
  void clear_timestamp() ;
  [[nodiscard]] const ::google::protobuf::Timestamp& timestamp() const;
  [[nodiscard]] ::google::protobuf::Timestamp* PROTOBUF_NULLABLE release_timestamp();
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL mutable_timestamp();
  void set_allocated_timestamp(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_timestamp(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  ::google::protobuf::Timestamp* PROTOBUF_NULLABLE unsafe_arena_release_timestamp();

  private:
  const ::google::protobuf::Timestamp& _internal_timestamp() const;
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL _internal_mutable_timestamp();

  public:
  // @@protoc_insertion_point(class_scope:agent.v1.CancelPendingRequest)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<2, 3,
                          1, 56,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const CancelPendingRequest& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr request_id_;
    ::google::protobuf::internal::ArenaStringPtr trace_id_;
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE timestamp_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_proto_2fagent_2fv1_2fcmds_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED AgentCommandResponse final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:agent.v1.AgentCommandResponse) */ {
 public:
  inline AgentCommandResponse() : AgentCommandResponse(nullptr) {}
  ~AgentCommandResponse() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(AgentCommandResponse* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(AgentCommandResponse));
  }
#endif

  template <typename = void>
  explicit constexpr AgentCommandResponse(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline AgentCommandResponse(const AgentCommandResponse& from) : AgentCommandResponse(nullptr, from) {}
  inline AgentCommandResponse(AgentCommandResponse&& from) noexcept : AgentCommandResponse(nullptr, ::std::move(from)) {}
  inline AgentCommandResponse& operator=(const AgentCommandResponse& from) {
    CopyFrom(from);
    return *this;
  }
  inline AgentCommandResponse& operator=(AgentCommandResponse&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const AgentCommandResponse& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<AgentCommandResponse>(&AgentCommandResponse_globals_);
  }
  static constexpr int kIndexInFileMessages = 3;
  friend void swap(AgentCommandResponse& a, AgentCommandResponse& b) { a.Swap(&b); }
  inline void Swap(AgentCommandResponse* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(AgentCommandResponse* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] AgentCommandResponse* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<AgentCommandResponse>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const AgentCommandResponse& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const AgentCommandResponse& from) { AgentCommandResponse::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(AgentCommandResponse* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "agent.v1.AgentCommandResponse"; }

  explicit AgentCommandResponse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  AgentCommandResponse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const AgentCommandResponse& from);
  AgentCommandResponse(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, AgentCommandResponse&& from) noexcept
      : AgentCommandResponse(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kMessageIdFieldNumber = 1,
    kResponseFieldNumber = 3,
    kTimestampFieldNumber = 4,
    kReceivedAtFieldNumber = 5,
    kExecTimeFieldNumber = 6,
    kStatusCodeFieldNumber = 2,
  };
  // string message_id = 1;
  void clear_message_id() ;
  [[nodiscard]] const ::std::string& message_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_message_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_message_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_message_id();
  void set_allocated_message_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_message_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_message_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_message_id();

  public:
  // bytes response = 3;
  void clear_response() ;
  [[nodiscard]] const ::std::string& response() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_response(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_response();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_response();
  void set_allocated_response(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_response() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_response(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_response();

  public:
  // .google.protobuf.Timestamp timestamp = 4;
  [[nodiscard]] bool has_timestamp()
      const;
  void clear_timestamp() ;
  [[nodiscard]] const ::google::protobuf::Timestamp& timestamp() const;
  [[nodiscard]] ::google::protobuf::Timestamp* PROTOBUF_NULLABLE release_timestamp();
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL mutable_timestamp();
  void set_allocated_timestamp(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_timestamp(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  ::google::protobuf::Timestamp* PROTOBUF_NULLABLE unsafe_arena_release_timestamp();

  private:
  const ::google::protobuf::Timestamp& _internal_timestamp() const;
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL _internal_mutable_timestamp();

  public:
  // .google.protobuf.Timestamp received_at = 5;
  [[nodiscard]] bool has_received_at()
      const;
  void clear_received_at() ;
  [[nodiscard]] const ::google::protobuf::Timestamp& received_at() const;
  [[nodiscard]] ::google::protobuf::Timestamp* PROTOBUF_NULLABLE release_received_at();
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL mutable_received_at();
  void set_allocated_received_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_received_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  ::google::protobuf::Timestamp* PROTOBUF_NULLABLE unsafe_arena_release_received_at();

  private:
  const ::google::protobuf::Timestamp& _internal_received_at() const;
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL _internal_mutable_received_at();

  public:
  // uint64 exec_time = 6;
  void clear_exec_time() ;
  [[nodiscard]] ::uint64_t exec_time() const;
  void set_exec_time(::uint64_t value);

  private:
  ::uint64_t _internal_exec_time() const;
  void _internal_set_exec_time(::uint64_t value);

  public:
  // uint32 status_code = 2;
  void clear_status_code() ;
  [[nodiscard]] ::uint32_t status_code() const;
  void set_status_code(::uint32_t value);

  private:
  ::uint32_t _internal_status_code() const;
  void _internal_set_status_code(::uint32_t value);

  public:
  // @@protoc_insertion_point(class_scope:agent.v1.AgentCommandResponse)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<3, 6,
                          2, 48,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const AgentCommandResponse& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr message_id_;
    ::google::protobuf::internal::ArenaStringPtr response_;
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE timestamp_;
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE received_at_;
    ::uint64_t exec_time_;
    ::uint32_t status_code_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_proto_2fagent_2fv1_2fcmds_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED AgentCommandAck final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:agent.v1.AgentCommandAck) */ {
 public:
  inline AgentCommandAck() : AgentCommandAck(nullptr) {}
  ~AgentCommandAck() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(AgentCommandAck* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(AgentCommandAck));
  }
#endif

  template <typename = void>
  explicit constexpr AgentCommandAck(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline AgentCommandAck(const AgentCommandAck& from) : AgentCommandAck(nullptr, from) {}
  inline AgentCommandAck(AgentCommandAck&& from) noexcept : AgentCommandAck(nullptr, ::std::move(from)) {}
  inline AgentCommandAck& operator=(const AgentCommandAck& from) {
    CopyFrom(from);
    return *this;
  }
  inline AgentCommandAck& operator=(AgentCommandAck&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const AgentCommandAck& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<AgentCommandAck>(&AgentCommandAck_globals_);
  }
  static constexpr int kIndexInFileMessages = 2;
  friend void swap(AgentCommandAck& a, AgentCommandAck& b) { a.Swap(&b); }
  inline void Swap(AgentCommandAck* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(AgentCommandAck* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] AgentCommandAck* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<AgentCommandAck>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const AgentCommandAck& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const AgentCommandAck& from) { AgentCommandAck::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(AgentCommandAck* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "agent.v1.AgentCommandAck"; }

  explicit AgentCommandAck(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  AgentCommandAck(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const AgentCommandAck& from);
  AgentCommandAck(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, AgentCommandAck&& from) noexcept
      : AgentCommandAck(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kMessageIdFieldNumber = 1,
    kCreatedAtFieldNumber = 2,
    kProgressPercentFieldNumber = 3,
  };
  // string message_id = 1;
  void clear_message_id() ;
  [[nodiscard]] const ::std::string& message_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_message_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_message_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_message_id();
  void set_allocated_message_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_message_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_message_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_message_id();

  public:
  // .google.protobuf.Timestamp created_at = 2;
  [[nodiscard]] bool has_created_at()
      const;
  void clear_created_at() ;
  [[nodiscard]] const ::google::protobuf::Timestamp& created_at() const;
  [[nodiscard]] ::google::protobuf::Timestamp* PROTOBUF_NULLABLE release_created_at();
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL mutable_created_at();
  void set_allocated_created_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_created_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  ::google::protobuf::Timestamp* PROTOBUF_NULLABLE unsafe_arena_release_created_at();

  private:
  const ::google::protobuf::Timestamp& _internal_created_at() const;
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL _internal_mutable_created_at();

  public:
  // uint32 progress_percent = 3;
  void clear_progress_percent() ;
  [[nodiscard]] ::uint32_t progress_percent() const;
  void set_progress_percent(::uint32_t value);

  private:
  ::uint32_t _internal_progress_percent() const;
  void _internal_set_progress_percent(::uint32_t value);

  public:
  // @@protoc_insertion_point(class_scope:agent.v1.AgentCommandAck)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<2, 3,
                          1, 43,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const AgentCommandAck& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr message_id_;
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE created_at_;
    ::uint32_t progress_percent_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_proto_2fagent_2fv1_2fcmds_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// CancelPendingRequest

// string request_id = 1;
inline void CancelPendingRequest::clear_request_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.request_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& CancelPendingRequest::request_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.CancelPendingRequest.request_id)
  return _internal_request_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void CancelPendingRequest::set_request_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.request_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.CancelPendingRequest.request_id)
}
inline ::std::string* PROTOBUF_NONNULL CancelPendingRequest::mutable_request_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_request_id();
  // @@protoc_insertion_point(field_mutable:agent.v1.CancelPendingRequest.request_id)
  return _s;
}
inline const ::std::string& CancelPendingRequest::_internal_request_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.request_id_.Get();
}
inline void CancelPendingRequest::_internal_set_request_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.request_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL CancelPendingRequest::_internal_mutable_request_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.request_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE CancelPendingRequest::release_request_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.CancelPendingRequest.request_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.request_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.request_id_.Set("", GetArena());
  }
  return released;
}
inline void CancelPendingRequest::set_allocated_request_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.request_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.request_id_.IsDefault()) {
    _impl_.request_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.CancelPendingRequest.request_id)
}

// .google.protobuf.Timestamp timestamp = 2;
inline bool CancelPendingRequest::has_timestamp() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000004U);
  PROTOBUF_ASSUME(!value || _impl_.timestamp_ != nullptr);
  return value;
}
inline const ::google::protobuf::Timestamp& CancelPendingRequest::_internal_timestamp() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::google::protobuf::Timestamp* p = _impl_.timestamp_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::google::protobuf::Timestamp>(&::google::protobuf::Timestamp_globals_);
}
inline const ::google::protobuf::Timestamp& CancelPendingRequest::timestamp() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.CancelPendingRequest.timestamp)
  return _internal_timestamp();
}
inline void CancelPendingRequest::unsafe_arena_set_allocated_timestamp(
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.timestamp_);
  }
  _impl_.timestamp_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:agent.v1.CancelPendingRequest.timestamp)
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE CancelPendingRequest::release_timestamp() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::google::protobuf::Timestamp* released = _impl_.timestamp_;
  _impl_.timestamp_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE CancelPendingRequest::unsafe_arena_release_timestamp() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.CancelPendingRequest.timestamp)

  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::google::protobuf::Timestamp* temp = _impl_.timestamp_;
  _impl_.timestamp_ = nullptr;
  return temp;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL CancelPendingRequest::_internal_mutable_timestamp() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.timestamp_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::google::protobuf::Timestamp>(GetArena());
    _impl_.timestamp_ = reinterpret_cast<::google::protobuf::Timestamp*>(p);
  }
  return _impl_.timestamp_;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL CancelPendingRequest::mutable_timestamp()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::google::protobuf::Timestamp* _msg = _internal_mutable_timestamp();
  // @@protoc_insertion_point(field_mutable:agent.v1.CancelPendingRequest.timestamp)
  return _msg;
}
inline void CancelPendingRequest::set_allocated_timestamp(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.timestamp_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }

  _impl_.timestamp_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  // @@protoc_insertion_point(field_set_allocated:agent.v1.CancelPendingRequest.timestamp)
}

// string trace_id = 3;
inline void CancelPendingRequest::clear_trace_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.trace_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& CancelPendingRequest::trace_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.CancelPendingRequest.trace_id)
  return _internal_trace_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void CancelPendingRequest::set_trace_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.trace_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.CancelPendingRequest.trace_id)
}
inline ::std::string* PROTOBUF_NONNULL CancelPendingRequest::mutable_trace_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_trace_id();
  // @@protoc_insertion_point(field_mutable:agent.v1.CancelPendingRequest.trace_id)
  return _s;
}
inline const ::std::string& CancelPendingRequest::_internal_trace_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.trace_id_.Get();
}
inline void CancelPendingRequest::_internal_set_trace_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.trace_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL CancelPendingRequest::_internal_mutable_trace_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.trace_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE CancelPendingRequest::release_trace_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.CancelPendingRequest.trace_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.trace_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.trace_id_.Set("", GetArena());
  }
  return released;
}
inline void CancelPendingRequest::set_allocated_trace_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.trace_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.trace_id_.IsDefault()) {
    _impl_.trace_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.CancelPendingRequest.trace_id)
}

// -------------------------------------------------------------------

// AgentCommand

// string callback_topic = 1;
inline void AgentCommand::clear_callback_topic() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.callback_topic_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& AgentCommand::callback_topic() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommand.callback_topic)
  return _internal_callback_topic();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void AgentCommand::set_callback_topic(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.callback_topic_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommand.callback_topic)
}
inline ::std::string* PROTOBUF_NONNULL AgentCommand::mutable_callback_topic()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_callback_topic();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommand.callback_topic)
  return _s;
}
inline const ::std::string& AgentCommand::_internal_callback_topic() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.callback_topic_.Get();
}
inline void AgentCommand::_internal_set_callback_topic(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.callback_topic_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL AgentCommand::_internal_mutable_callback_topic() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.callback_topic_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE AgentCommand::release_callback_topic() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommand.callback_topic)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.callback_topic_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.callback_topic_.Set("", GetArena());
  }
  return released;
}
inline void AgentCommand::set_allocated_callback_topic(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.callback_topic_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.callback_topic_.IsDefault()) {
    _impl_.callback_topic_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommand.callback_topic)
}

// string ack_topic = 2;
inline void AgentCommand::clear_ack_topic() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.ack_topic_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& AgentCommand::ack_topic() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommand.ack_topic)
  return _internal_ack_topic();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void AgentCommand::set_ack_topic(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.ack_topic_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommand.ack_topic)
}
inline ::std::string* PROTOBUF_NONNULL AgentCommand::mutable_ack_topic()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_ack_topic();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommand.ack_topic)
  return _s;
}
inline const ::std::string& AgentCommand::_internal_ack_topic() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.ack_topic_.Get();
}
inline void AgentCommand::_internal_set_ack_topic(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.ack_topic_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL AgentCommand::_internal_mutable_ack_topic() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.ack_topic_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE AgentCommand::release_ack_topic() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommand.ack_topic)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.ack_topic_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.ack_topic_.Set("", GetArena());
  }
  return released;
}
inline void AgentCommand::set_allocated_ack_topic(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.ack_topic_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.ack_topic_.IsDefault()) {
    _impl_.ack_topic_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommand.ack_topic)
}

// string message_id = 3;
inline void AgentCommand::clear_message_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.message_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline const ::std::string& AgentCommand::message_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommand.message_id)
  return _internal_message_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void AgentCommand::set_message_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  _impl_.message_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommand.message_id)
}
inline ::std::string* PROTOBUF_NONNULL AgentCommand::mutable_message_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::std::string* _s = _internal_mutable_message_id();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommand.message_id)
  return _s;
}
inline const ::std::string& AgentCommand::_internal_message_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.message_id_.Get();
}
inline void AgentCommand::_internal_set_message_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.message_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL AgentCommand::_internal_mutable_message_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.message_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE AgentCommand::release_message_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommand.message_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000004U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  auto* released = _impl_.message_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.message_id_.Set("", GetArena());
  }
  return released;
}
inline void AgentCommand::set_allocated_message_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  _impl_.message_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.message_id_.IsDefault()) {
    _impl_.message_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommand.message_id)
}

// uint64 timeout = 4;
inline void AgentCommand::clear_timeout() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.timeout_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
}
inline ::uint64_t AgentCommand::timeout() const {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommand.timeout)
  return _internal_timeout();
}
inline void AgentCommand::set_timeout(::uint64_t value) {
  _internal_set_timeout(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommand.timeout)
}
inline ::uint64_t AgentCommand::_internal_timeout() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.timeout_;
}
inline void AgentCommand::_internal_set_timeout(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.timeout_ = value;
}

// uint64 ack_timeout = 5;
inline void AgentCommand::clear_ack_timeout() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.ack_timeout_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
}
inline ::uint64_t AgentCommand::ack_timeout() const {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommand.ack_timeout)
  return _internal_ack_timeout();
}
inline void AgentCommand::set_ack_timeout(::uint64_t value) {
  _internal_set_ack_timeout(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommand.ack_timeout)
}
inline ::uint64_t AgentCommand::_internal_ack_timeout() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.ack_timeout_;
}
inline void AgentCommand::_internal_set_ack_timeout(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.ack_timeout_ = value;
}

// string resource = 6;
inline void AgentCommand::clear_resource() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.resource_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline const ::std::string& AgentCommand::resource() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommand.resource)
  return _internal_resource();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void AgentCommand::set_resource(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  _impl_.resource_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommand.resource)
}
inline ::std::string* PROTOBUF_NONNULL AgentCommand::mutable_resource()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  ::std::string* _s = _internal_mutable_resource();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommand.resource)
  return _s;
}
inline const ::std::string& AgentCommand::_internal_resource() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.resource_.Get();
}
inline void AgentCommand::_internal_set_resource(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.resource_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL AgentCommand::_internal_mutable_resource() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.resource_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE AgentCommand::release_resource() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommand.resource)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000008U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  auto* released = _impl_.resource_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.resource_.Set("", GetArena());
  }
  return released;
}
inline void AgentCommand::set_allocated_resource(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  }
  _impl_.resource_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.resource_.IsDefault()) {
    _impl_.resource_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommand.resource)
}

// -------------------------------------------------------------------

// AgentCommandAck

// string message_id = 1;
inline void AgentCommandAck::clear_message_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.message_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& AgentCommandAck::message_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommandAck.message_id)
  return _internal_message_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void AgentCommandAck::set_message_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.message_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommandAck.message_id)
}
inline ::std::string* PROTOBUF_NONNULL AgentCommandAck::mutable_message_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_message_id();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommandAck.message_id)
  return _s;
}
inline const ::std::string& AgentCommandAck::_internal_message_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.message_id_.Get();
}
inline void AgentCommandAck::_internal_set_message_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.message_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL AgentCommandAck::_internal_mutable_message_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.message_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE AgentCommandAck::release_message_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommandAck.message_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.message_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.message_id_.Set("", GetArena());
  }
  return released;
}
inline void AgentCommandAck::set_allocated_message_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.message_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.message_id_.IsDefault()) {
    _impl_.message_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommandAck.message_id)
}

// .google.protobuf.Timestamp created_at = 2;
inline bool AgentCommandAck::has_created_at() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000002U);
  PROTOBUF_ASSUME(!value || _impl_.created_at_ != nullptr);
  return value;
}
inline const ::google::protobuf::Timestamp& AgentCommandAck::_internal_created_at() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::google::protobuf::Timestamp* p = _impl_.created_at_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::google::protobuf::Timestamp>(&::google::protobuf::Timestamp_globals_);
}
inline const ::google::protobuf::Timestamp& AgentCommandAck::created_at() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommandAck.created_at)
  return _internal_created_at();
}
inline void AgentCommandAck::unsafe_arena_set_allocated_created_at(
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.created_at_);
  }
  _impl_.created_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:agent.v1.AgentCommandAck.created_at)
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE AgentCommandAck::release_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::google::protobuf::Timestamp* released = _impl_.created_at_;
  _impl_.created_at_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE AgentCommandAck::unsafe_arena_release_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommandAck.created_at)

  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::google::protobuf::Timestamp* temp = _impl_.created_at_;
  _impl_.created_at_ = nullptr;
  return temp;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL AgentCommandAck::_internal_mutable_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.created_at_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::google::protobuf::Timestamp>(GetArena());
    _impl_.created_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(p);
  }
  return _impl_.created_at_;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL AgentCommandAck::mutable_created_at()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::google::protobuf::Timestamp* _msg = _internal_mutable_created_at();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommandAck.created_at)
  return _msg;
}
inline void AgentCommandAck::set_allocated_created_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.created_at_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }

  _impl_.created_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommandAck.created_at)
}

// uint32 progress_percent = 3;
inline void AgentCommandAck::clear_progress_percent() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.progress_percent_ = 0u;
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline ::uint32_t AgentCommandAck::progress_percent() const {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommandAck.progress_percent)
  return _internal_progress_percent();
}
inline void AgentCommandAck::set_progress_percent(::uint32_t value) {
  _internal_set_progress_percent(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommandAck.progress_percent)
}
inline ::uint32_t AgentCommandAck::_internal_progress_percent() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.progress_percent_;
}
inline void AgentCommandAck::_internal_set_progress_percent(::uint32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.progress_percent_ = value;
}

// -------------------------------------------------------------------

// AgentCommandResponse

// string message_id = 1;
inline void AgentCommandResponse::clear_message_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.message_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& AgentCommandResponse::message_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommandResponse.message_id)
  return _internal_message_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void AgentCommandResponse::set_message_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.message_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommandResponse.message_id)
}
inline ::std::string* PROTOBUF_NONNULL AgentCommandResponse::mutable_message_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_message_id();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommandResponse.message_id)
  return _s;
}
inline const ::std::string& AgentCommandResponse::_internal_message_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.message_id_.Get();
}
inline void AgentCommandResponse::_internal_set_message_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.message_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL AgentCommandResponse::_internal_mutable_message_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.message_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE AgentCommandResponse::release_message_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommandResponse.message_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.message_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.message_id_.Set("", GetArena());
  }
  return released;
}
inline void AgentCommandResponse::set_allocated_message_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.message_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.message_id_.IsDefault()) {
    _impl_.message_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommandResponse.message_id)
}

// uint32 status_code = 2;
inline void AgentCommandResponse::clear_status_code() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.status_code_ = 0u;
  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
}
inline ::uint32_t AgentCommandResponse::status_code() const {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommandResponse.status_code)
  return _internal_status_code();
}
inline void AgentCommandResponse::set_status_code(::uint32_t value) {
  _internal_set_status_code(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommandResponse.status_code)
}
inline ::uint32_t AgentCommandResponse::_internal_status_code() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.status_code_;
}
inline void AgentCommandResponse::_internal_set_status_code(::uint32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.status_code_ = value;
}

// bytes response = 3;
inline void AgentCommandResponse::clear_response() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.response_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& AgentCommandResponse::response() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommandResponse.response)
  return _internal_response();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void AgentCommandResponse::set_response(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.response_.SetBytes(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommandResponse.response)
}
inline ::std::string* PROTOBUF_NONNULL AgentCommandResponse::mutable_response()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_response();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommandResponse.response)
  return _s;
}
inline const ::std::string& AgentCommandResponse::_internal_response() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.response_.Get();
}
inline void AgentCommandResponse::_internal_set_response(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.response_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL AgentCommandResponse::_internal_mutable_response() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.response_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE AgentCommandResponse::release_response() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommandResponse.response)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.response_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.response_.Set("", GetArena());
  }
  return released;
}
inline void AgentCommandResponse::set_allocated_response(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.response_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.response_.IsDefault()) {
    _impl_.response_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommandResponse.response)
}

// .google.protobuf.Timestamp timestamp = 4;
inline bool AgentCommandResponse::has_timestamp() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000004U);
  PROTOBUF_ASSUME(!value || _impl_.timestamp_ != nullptr);
  return value;
}
inline const ::google::protobuf::Timestamp& AgentCommandResponse::_internal_timestamp() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::google::protobuf::Timestamp* p = _impl_.timestamp_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::google::protobuf::Timestamp>(&::google::protobuf::Timestamp_globals_);
}
inline const ::google::protobuf::Timestamp& AgentCommandResponse::timestamp() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommandResponse.timestamp)
  return _internal_timestamp();
}
inline void AgentCommandResponse::unsafe_arena_set_allocated_timestamp(
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.timestamp_);
  }
  _impl_.timestamp_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:agent.v1.AgentCommandResponse.timestamp)
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE AgentCommandResponse::release_timestamp() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::google::protobuf::Timestamp* released = _impl_.timestamp_;
  _impl_.timestamp_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE AgentCommandResponse::unsafe_arena_release_timestamp() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommandResponse.timestamp)

  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::google::protobuf::Timestamp* temp = _impl_.timestamp_;
  _impl_.timestamp_ = nullptr;
  return temp;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL AgentCommandResponse::_internal_mutable_timestamp() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.timestamp_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::google::protobuf::Timestamp>(GetArena());
    _impl_.timestamp_ = reinterpret_cast<::google::protobuf::Timestamp*>(p);
  }
  return _impl_.timestamp_;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL AgentCommandResponse::mutable_timestamp()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::google::protobuf::Timestamp* _msg = _internal_mutable_timestamp();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommandResponse.timestamp)
  return _msg;
}
inline void AgentCommandResponse::set_allocated_timestamp(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.timestamp_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }

  _impl_.timestamp_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommandResponse.timestamp)
}

// .google.protobuf.Timestamp received_at = 5;
inline bool AgentCommandResponse::has_received_at() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000008U);
  PROTOBUF_ASSUME(!value || _impl_.received_at_ != nullptr);
  return value;
}
inline const ::google::protobuf::Timestamp& AgentCommandResponse::_internal_received_at() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::google::protobuf::Timestamp* p = _impl_.received_at_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::google::protobuf::Timestamp>(&::google::protobuf::Timestamp_globals_);
}
inline const ::google::protobuf::Timestamp& AgentCommandResponse::received_at() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommandResponse.received_at)
  return _internal_received_at();
}
inline void AgentCommandResponse::unsafe_arena_set_allocated_received_at(
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.received_at_);
  }
  _impl_.received_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:agent.v1.AgentCommandResponse.received_at)
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE AgentCommandResponse::release_received_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  ::google::protobuf::Timestamp* released = _impl_.received_at_;
  _impl_.received_at_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE AgentCommandResponse::unsafe_arena_release_received_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.AgentCommandResponse.received_at)

  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  ::google::protobuf::Timestamp* temp = _impl_.received_at_;
  _impl_.received_at_ = nullptr;
  return temp;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL AgentCommandResponse::_internal_mutable_received_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.received_at_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::google::protobuf::Timestamp>(GetArena());
    _impl_.received_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(p);
  }
  return _impl_.received_at_;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL AgentCommandResponse::mutable_received_at()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  ::google::protobuf::Timestamp* _msg = _internal_mutable_received_at();
  // @@protoc_insertion_point(field_mutable:agent.v1.AgentCommandResponse.received_at)
  return _msg;
}
inline void AgentCommandResponse::set_allocated_received_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.received_at_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  }

  _impl_.received_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  // @@protoc_insertion_point(field_set_allocated:agent.v1.AgentCommandResponse.received_at)
}

// uint64 exec_time = 6;
inline void AgentCommandResponse::clear_exec_time() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.exec_time_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
}
inline ::uint64_t AgentCommandResponse::exec_time() const {
  // @@protoc_insertion_point(field_get:agent.v1.AgentCommandResponse.exec_time)
  return _internal_exec_time();
}
inline void AgentCommandResponse::set_exec_time(::uint64_t value) {
  _internal_set_exec_time(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  // @@protoc_insertion_point(field_set:agent.v1.AgentCommandResponse.exec_time)
}
inline ::uint64_t AgentCommandResponse::_internal_exec_time() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.exec_time_;
}
inline void AgentCommandResponse::_internal_set_exec_time(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.exec_time_ = value;
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace agent


// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // proto_2fagent_2fv1_2fcmds_2eproto_2epb_2eh
