// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: proto/context/v1/context.proto
// Protobuf C++ Version: 7.35.1

#ifndef proto_2fcontext_2fv1_2fcontext_2eproto_2epb_2eh
#define proto_2fcontext_2fv1_2fcontext_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/unknown_field_set.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_proto_2fcontext_2fv1_2fcontext_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_proto_2fcontext_2fv1_2fcontext_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_proto_2fcontext_2fv1_2fcontext_2eproto;
}  // extern "C"
namespace context {
namespace v1 {
class ContextUpdated;
struct ContextUpdatedGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern ContextUpdatedGlobalsTypeInternal ContextUpdated_globals_;
extern const ::google::protobuf::internal::ClassDataFull ContextUpdated_class_data_;
#else
extern const ContextUpdatedGlobalsTypeInternal ContextUpdated_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class ContextsSnapshot;
struct ContextsSnapshotGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern ContextsSnapshotGlobalsTypeInternal ContextsSnapshot_globals_;
extern const ::google::protobuf::internal::ClassDataFull ContextsSnapshot_class_data_;
#else
extern const ContextsSnapshotGlobalsTypeInternal ContextsSnapshot_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class ContextsUpdated;
struct ContextsUpdatedGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern ContextsUpdatedGlobalsTypeInternal ContextsUpdated_globals_;
extern const ::google::protobuf::internal::ClassDataFull ContextsUpdated_class_data_;
#else
extern const ContextsUpdatedGlobalsTypeInternal ContextsUpdated_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace v1
}  // namespace context
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google

namespace context {
namespace v1 {

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED ContextUpdated final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:context.v1.ContextUpdated) */ {
 public:
  inline ContextUpdated() : ContextUpdated(nullptr) {}
  ~ContextUpdated() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(ContextUpdated* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(ContextUpdated));
  }
#endif

  template <typename = void>
  explicit constexpr ContextUpdated(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline ContextUpdated(const ContextUpdated& from) : ContextUpdated(nullptr, from) {}
  inline ContextUpdated(ContextUpdated&& from) noexcept : ContextUpdated(nullptr, ::std::move(from)) {}
  inline ContextUpdated& operator=(const ContextUpdated& from) {
    CopyFrom(from);
    return *this;
  }
  inline ContextUpdated& operator=(ContextUpdated&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const ContextUpdated& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<ContextUpdated>(&ContextUpdated_globals_);
  }
  static constexpr int kIndexInFileMessages = 1;
  friend void swap(ContextUpdated& a, ContextUpdated& b) { a.Swap(&b); }
  inline void Swap(ContextUpdated* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(ContextUpdated* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] ContextUpdated* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<ContextUpdated>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const ContextUpdated& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const ContextUpdated& from) { ContextUpdated::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(ContextUpdated* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "context.v1.ContextUpdated"; }

  explicit ContextUpdated(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  ContextUpdated(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ContextUpdated& from);
  ContextUpdated(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, ContextUpdated&& from) noexcept
      : ContextUpdated(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kIdFieldNumber = 1,
    kTitleFieldNumber = 6,
    kChartTypeFieldNumber = 8,
    kUnitsFieldNumber = 9,
    kFamilyFieldNumber = 10,
    kVersionFieldNumber = 2,
    kFirstEntryFieldNumber = 3,
    kLastEntryFieldNumber = 4,
    kPriorityFieldNumber = 7,
    kDeletedFieldNumber = 5,
  };
  // string id = 1;
  void clear_id() ;
  [[nodiscard]] const ::std::string& id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_id();
  void set_allocated_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_id();

  public:
  // string title = 6;
  void clear_title() ;
  [[nodiscard]] const ::std::string& title() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_title(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_title();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_title();
  void set_allocated_title(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_title() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_title(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_title();

  public:
  // string chart_type = 8;
  void clear_chart_type() ;
  [[nodiscard]] const ::std::string& chart_type() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_chart_type(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_chart_type();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_chart_type();
  void set_allocated_chart_type(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_chart_type() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_chart_type(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_chart_type();

  public:
  // string units = 9;
  void clear_units() ;
  [[nodiscard]] const ::std::string& units() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_units(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_units();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_units();
  void set_allocated_units(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_units() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_units(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_units();

  public:
  // string family = 10;
  void clear_family() ;
  [[nodiscard]] const ::std::string& family() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_family(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_family();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_family();
  void set_allocated_family(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_family() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_family(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_family();

  public:
  // uint64 version = 2;
  void clear_version() ;
  [[nodiscard]] ::uint64_t version() const;
  void set_version(::uint64_t value);

  private:
  ::uint64_t _internal_version() const;
  void _internal_set_version(::uint64_t value);

  public:
  // uint64 first_entry = 3;
  void clear_first_entry() ;
  [[nodiscard]] ::uint64_t first_entry() const;
  void set_first_entry(::uint64_t value);

  private:
  ::uint64_t _internal_first_entry() const;
  void _internal_set_first_entry(::uint64_t value);

  public:
  // uint64 last_entry = 4;
  void clear_last_entry() ;
  [[nodiscard]] ::uint64_t last_entry() const;
  void set_last_entry(::uint64_t value);

  private:
  ::uint64_t _internal_last_entry() const;
  void _internal_set_last_entry(::uint64_t value);

  public:
  // uint64 priority = 7;
  void clear_priority() ;
  [[nodiscard]] ::uint64_t priority() const;
  void set_priority(::uint64_t value);

  private:
  ::uint64_t _internal_priority() const;
  void _internal_set_priority(::uint64_t value);

  public:
  // bool deleted = 5;
  void clear_deleted() ;
  [[nodiscard]] bool deleted() const;
  void set_deleted(bool value);

  private:
  bool _internal_deleted() const;
  void _internal_set_deleted(bool value);

  public:
  // @@protoc_insertion_point(class_scope:context.v1.ContextUpdated)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<4, 10,
                          0, 70,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const ContextUpdated& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr id_;
    ::google::protobuf::internal::ArenaStringPtr title_;
    ::google::protobuf::internal::ArenaStringPtr chart_type_;
    ::google::protobuf::internal::ArenaStringPtr units_;
    ::google::protobuf::internal::ArenaStringPtr family_;
    ::uint64_t version_;
    ::uint64_t first_entry_;
    ::uint64_t last_entry_;
    ::uint64_t priority_;
    bool deleted_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_proto_2fcontext_2fv1_2fcontext_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED ContextsUpdated final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:context.v1.ContextsUpdated) */ {
 public:
  inline ContextsUpdated() : ContextsUpdated(nullptr) {}
  ~ContextsUpdated() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(ContextsUpdated* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(ContextsUpdated));
  }
#endif

  template <typename = void>
  explicit constexpr ContextsUpdated(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline ContextsUpdated(const ContextsUpdated& from) : ContextsUpdated(nullptr, from) {}
  inline ContextsUpdated(ContextsUpdated&& from) noexcept : ContextsUpdated(nullptr, ::std::move(from)) {}
  inline ContextsUpdated& operator=(const ContextsUpdated& from) {
    CopyFrom(from);
    return *this;
  }
  inline ContextsUpdated& operator=(ContextsUpdated&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const ContextsUpdated& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<ContextsUpdated>(&ContextsUpdated_globals_);
  }
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(ContextsUpdated& a, ContextsUpdated& b) { a.Swap(&b); }
  inline void Swap(ContextsUpdated* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(ContextsUpdated* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] ContextsUpdated* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<ContextsUpdated>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const ContextsUpdated& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const ContextsUpdated& from) { ContextsUpdated::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(ContextsUpdated* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "context.v1.ContextsUpdated"; }

  explicit ContextsUpdated(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  ContextsUpdated(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ContextsUpdated& from);
  ContextsUpdated(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, ContextsUpdated&& from) noexcept
      : ContextsUpdated(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kContextUpdatesFieldNumber = 1,
    kClaimIdFieldNumber = 2,
    kNodeIdFieldNumber = 3,
    kVersionHashFieldNumber = 4,
    kCreatedAtFieldNumber = 5,
  };
  // repeated .context.v1.ContextUpdated contextUpdates = 1;
  [[nodiscard]] int contextupdates_size()
      const;
  private:
  int _internal_contextupdates_size() const;

  public:
  void clear_contextupdates() ;
  [[nodiscard]] const ::context::v1::ContextUpdated& contextupdates(int index) const;
  [[nodiscard]] ::context::v1::ContextUpdated* PROTOBUF_NONNULL mutable_contextupdates(int index);
  ::context::v1::ContextUpdated* PROTOBUF_NONNULL add_contextupdates();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>&
  contextupdates() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>* PROTOBUF_NONNULL
  mutable_contextupdates();

  private:
  const ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>& _internal_contextupdates() const;
  ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>* PROTOBUF_NONNULL _internal_mutable_contextupdates();

  public:
  // string claim_id = 2;
  void clear_claim_id() ;
  [[nodiscard]] const ::std::string& claim_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_claim_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_claim_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_claim_id();
  void set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_claim_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_claim_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_claim_id();

  public:
  // string node_id = 3;
  void clear_node_id() ;
  [[nodiscard]] const ::std::string& node_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_node_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_node_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_node_id();
  void set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_node_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_node_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_node_id();

  public:
  // uint64 version_hash = 4;
  void clear_version_hash() ;
  [[nodiscard]] ::uint64_t version_hash() const;
  void set_version_hash(::uint64_t value);

  private:
  ::uint64_t _internal_version_hash() const;
  void _internal_set_version_hash(::uint64_t value);

  public:
  // uint64 created_at = 5;
  void clear_created_at() ;
  [[nodiscard]] ::uint64_t created_at() const;
  void set_created_at(::uint64_t value);

  private:
  ::uint64_t _internal_created_at() const;
  void _internal_set_created_at(::uint64_t value);

  public:
  // @@protoc_insertion_point(class_scope:context.v1.ContextsUpdated)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<3, 5,
                          1, 50,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const ContextsUpdated& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField< ::context::v1::ContextUpdated > contextupdates_;
    ::google::protobuf::internal::ArenaStringPtr claim_id_;
    ::google::protobuf::internal::ArenaStringPtr node_id_;
    ::uint64_t version_hash_;
    ::uint64_t created_at_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_proto_2fcontext_2fv1_2fcontext_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED ContextsSnapshot final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:context.v1.ContextsSnapshot) */ {
 public:
  inline ContextsSnapshot() : ContextsSnapshot(nullptr) {}
  ~ContextsSnapshot() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(ContextsSnapshot* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(ContextsSnapshot));
  }
#endif

  template <typename = void>
  explicit constexpr ContextsSnapshot(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline ContextsSnapshot(const ContextsSnapshot& from) : ContextsSnapshot(nullptr, from) {}
  inline ContextsSnapshot(ContextsSnapshot&& from) noexcept : ContextsSnapshot(nullptr, ::std::move(from)) {}
  inline ContextsSnapshot& operator=(const ContextsSnapshot& from) {
    CopyFrom(from);
    return *this;
  }
  inline ContextsSnapshot& operator=(ContextsSnapshot&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const ContextsSnapshot& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<ContextsSnapshot>(&ContextsSnapshot_globals_);
  }
  static constexpr int kIndexInFileMessages = 2;
  friend void swap(ContextsSnapshot& a, ContextsSnapshot& b) { a.Swap(&b); }
  inline void Swap(ContextsSnapshot* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(ContextsSnapshot* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] ContextsSnapshot* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<ContextsSnapshot>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const ContextsSnapshot& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const ContextsSnapshot& from) { ContextsSnapshot::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(ContextsSnapshot* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "context.v1.ContextsSnapshot"; }

  explicit ContextsSnapshot(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  ContextsSnapshot(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ContextsSnapshot& from);
  ContextsSnapshot(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, ContextsSnapshot&& from) noexcept
      : ContextsSnapshot(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kContextsFieldNumber = 1,
    kClaimIdFieldNumber = 2,
    kNodeIdFieldNumber = 3,
    kVersionFieldNumber = 4,
  };
  // repeated .context.v1.ContextUpdated contexts = 1;
  [[nodiscard]] int contexts_size()
      const;
  private:
  int _internal_contexts_size() const;

  public:
  void clear_contexts() ;
  [[nodiscard]] const ::context::v1::ContextUpdated& contexts(int index) const;
  [[nodiscard]] ::context::v1::ContextUpdated* PROTOBUF_NONNULL mutable_contexts(int index);
  ::context::v1::ContextUpdated* PROTOBUF_NONNULL add_contexts();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>&
  contexts() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>* PROTOBUF_NONNULL
  mutable_contexts();

  private:
  const ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>& _internal_contexts() const;
  ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>* PROTOBUF_NONNULL _internal_mutable_contexts();

  public:
  // string claim_id = 2;
  void clear_claim_id() ;
  [[nodiscard]] const ::std::string& claim_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_claim_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_claim_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_claim_id();
  void set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_claim_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_claim_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_claim_id();

  public:
  // string node_id = 3;
  void clear_node_id() ;
  [[nodiscard]] const ::std::string& node_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_node_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_node_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_node_id();
  void set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_node_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_node_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_node_id();

  public:
  // uint64 version = 4;
  void clear_version() ;
  [[nodiscard]] ::uint64_t version() const;
  void set_version(::uint64_t value);

  private:
  ::uint64_t _internal_version() const;
  void _internal_set_version(::uint64_t value);

  public:
  // @@protoc_insertion_point(class_scope:context.v1.ContextsSnapshot)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<2, 4,
                          1, 51,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const ContextsSnapshot& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField< ::context::v1::ContextUpdated > contexts_;
    ::google::protobuf::internal::ArenaStringPtr claim_id_;
    ::google::protobuf::internal::ArenaStringPtr node_id_;
    ::uint64_t version_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_proto_2fcontext_2fv1_2fcontext_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// ContextsUpdated

// repeated .context.v1.ContextUpdated contextUpdates = 1;
inline int ContextsUpdated::_internal_contextupdates_size() const {
  return _internal_contextupdates().size();
}
inline int ContextsUpdated::contextupdates_size() const {
  return _internal_contextupdates_size();
}
inline void ContextsUpdated::clear_contextupdates() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.contextupdates_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::context::v1::ContextUpdated& ContextsUpdated::contextupdates(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextsUpdated.contextUpdates)
  return _internal_contextupdates().Get(index);
}
inline ::context::v1::ContextUpdated* PROTOBUF_NONNULL ContextsUpdated::mutable_contextupdates(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:context.v1.ContextsUpdated.contextUpdates)
  return _internal_mutable_contextupdates()->Mutable(index);
}
inline ::context::v1::ContextUpdated* PROTOBUF_NONNULL ContextsUpdated::add_contextupdates()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::context::v1::ContextUpdated* _add =
      _internal_mutable_contextupdates()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:context.v1.ContextsUpdated.contextUpdates)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>& ContextsUpdated::contextupdates() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:context.v1.ContextsUpdated.contextUpdates)
  return _internal_contextupdates();
}
inline ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>* PROTOBUF_NONNULL
ContextsUpdated::mutable_contextupdates() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:context.v1.ContextsUpdated.contextUpdates)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_contextupdates();
}
inline const ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>&
ContextsUpdated::_internal_contextupdates() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.contextupdates_;
}
inline ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>* PROTOBUF_NONNULL
ContextsUpdated::_internal_mutable_contextupdates() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.contextupdates_;
}

// string claim_id = 2;
inline void ContextsUpdated::clear_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& ContextsUpdated::claim_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextsUpdated.claim_id)
  return _internal_claim_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextsUpdated::set_claim_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.claim_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextsUpdated.claim_id)
}
inline ::std::string* PROTOBUF_NONNULL ContextsUpdated::mutable_claim_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_claim_id();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextsUpdated.claim_id)
  return _s;
}
inline const ::std::string& ContextsUpdated::_internal_claim_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.claim_id_.Get();
}
inline void ContextsUpdated::_internal_set_claim_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextsUpdated::_internal_mutable_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.claim_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextsUpdated::release_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextsUpdated.claim_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.claim_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  return released;
}
inline void ContextsUpdated::set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.claim_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.claim_id_.IsDefault()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextsUpdated.claim_id)
}

// string node_id = 3;
inline void ContextsUpdated::clear_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline const ::std::string& ContextsUpdated::node_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextsUpdated.node_id)
  return _internal_node_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextsUpdated::set_node_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  _impl_.node_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextsUpdated.node_id)
}
inline ::std::string* PROTOBUF_NONNULL ContextsUpdated::mutable_node_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::std::string* _s = _internal_mutable_node_id();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextsUpdated.node_id)
  return _s;
}
inline const ::std::string& ContextsUpdated::_internal_node_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.node_id_.Get();
}
inline void ContextsUpdated::_internal_set_node_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextsUpdated::_internal_mutable_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.node_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextsUpdated::release_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextsUpdated.node_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000004U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  auto* released = _impl_.node_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.node_id_.Set("", GetArena());
  }
  return released;
}
inline void ContextsUpdated::set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  _impl_.node_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.node_id_.IsDefault()) {
    _impl_.node_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextsUpdated.node_id)
}

// uint64 version_hash = 4;
inline void ContextsUpdated::clear_version_hash() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_hash_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline ::uint64_t ContextsUpdated::version_hash() const {
  // @@protoc_insertion_point(field_get:context.v1.ContextsUpdated.version_hash)
  return _internal_version_hash();
}
inline void ContextsUpdated::set_version_hash(::uint64_t value) {
  _internal_set_version_hash(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_set:context.v1.ContextsUpdated.version_hash)
}
inline ::uint64_t ContextsUpdated::_internal_version_hash() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.version_hash_;
}
inline void ContextsUpdated::_internal_set_version_hash(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_hash_ = value;
}

// uint64 created_at = 5;
inline void ContextsUpdated::clear_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.created_at_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
}
inline ::uint64_t ContextsUpdated::created_at() const {
  // @@protoc_insertion_point(field_get:context.v1.ContextsUpdated.created_at)
  return _internal_created_at();
}
inline void ContextsUpdated::set_created_at(::uint64_t value) {
  _internal_set_created_at(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  // @@protoc_insertion_point(field_set:context.v1.ContextsUpdated.created_at)
}
inline ::uint64_t ContextsUpdated::_internal_created_at() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.created_at_;
}
inline void ContextsUpdated::_internal_set_created_at(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.created_at_ = value;
}

// -------------------------------------------------------------------

// ContextUpdated

// string id = 1;
inline void ContextUpdated::clear_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& ContextUpdated::id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.id)
  return _internal_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextUpdated::set_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.id)
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::mutable_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_id();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextUpdated.id)
  return _s;
}
inline const ::std::string& ContextUpdated::_internal_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.id_.Get();
}
inline void ContextUpdated::_internal_set_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::_internal_mutable_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextUpdated::release_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextUpdated.id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.id_.Set("", GetArena());
  }
  return released;
}
inline void ContextUpdated::set_allocated_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.id_.IsDefault()) {
    _impl_.id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextUpdated.id)
}

// uint64 version = 2;
inline void ContextUpdated::clear_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
}
inline ::uint64_t ContextUpdated::version() const {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.version)
  return _internal_version();
}
inline void ContextUpdated::set_version(::uint64_t value) {
  _internal_set_version(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.version)
}
inline ::uint64_t ContextUpdated::_internal_version() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.version_;
}
inline void ContextUpdated::_internal_set_version(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_ = value;
}

// uint64 first_entry = 3;
inline void ContextUpdated::clear_first_entry() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.first_entry_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000040U);
}
inline ::uint64_t ContextUpdated::first_entry() const {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.first_entry)
  return _internal_first_entry();
}
inline void ContextUpdated::set_first_entry(::uint64_t value) {
  _internal_set_first_entry(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000040U);
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.first_entry)
}
inline ::uint64_t ContextUpdated::_internal_first_entry() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.first_entry_;
}
inline void ContextUpdated::_internal_set_first_entry(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.first_entry_ = value;
}

// uint64 last_entry = 4;
inline void ContextUpdated::clear_last_entry() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.last_entry_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000080U);
}
inline ::uint64_t ContextUpdated::last_entry() const {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.last_entry)
  return _internal_last_entry();
}
inline void ContextUpdated::set_last_entry(::uint64_t value) {
  _internal_set_last_entry(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000080U);
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.last_entry)
}
inline ::uint64_t ContextUpdated::_internal_last_entry() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.last_entry_;
}
inline void ContextUpdated::_internal_set_last_entry(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.last_entry_ = value;
}

// bool deleted = 5;
inline void ContextUpdated::clear_deleted() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.deleted_ = false;
  ClearHasBit(_impl_._has_bits_[0], 0x00000200U);
}
inline bool ContextUpdated::deleted() const {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.deleted)
  return _internal_deleted();
}
inline void ContextUpdated::set_deleted(bool value) {
  _internal_set_deleted(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000200U);
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.deleted)
}
inline bool ContextUpdated::_internal_deleted() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.deleted_;
}
inline void ContextUpdated::_internal_set_deleted(bool value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.deleted_ = value;
}

// string title = 6;
inline void ContextUpdated::clear_title() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.title_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& ContextUpdated::title() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.title)
  return _internal_title();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextUpdated::set_title(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.title_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.title)
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::mutable_title()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_title();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextUpdated.title)
  return _s;
}
inline const ::std::string& ContextUpdated::_internal_title() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.title_.Get();
}
inline void ContextUpdated::_internal_set_title(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.title_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::_internal_mutable_title() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.title_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextUpdated::release_title() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextUpdated.title)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.title_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.title_.Set("", GetArena());
  }
  return released;
}
inline void ContextUpdated::set_allocated_title(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.title_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.title_.IsDefault()) {
    _impl_.title_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextUpdated.title)
}

// uint64 priority = 7;
inline void ContextUpdated::clear_priority() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.priority_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000100U);
}
inline ::uint64_t ContextUpdated::priority() const {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.priority)
  return _internal_priority();
}
inline void ContextUpdated::set_priority(::uint64_t value) {
  _internal_set_priority(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000100U);
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.priority)
}
inline ::uint64_t ContextUpdated::_internal_priority() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.priority_;
}
inline void ContextUpdated::_internal_set_priority(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.priority_ = value;
}

// string chart_type = 8;
inline void ContextUpdated::clear_chart_type() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.chart_type_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline const ::std::string& ContextUpdated::chart_type() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.chart_type)
  return _internal_chart_type();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextUpdated::set_chart_type(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  _impl_.chart_type_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.chart_type)
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::mutable_chart_type()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::std::string* _s = _internal_mutable_chart_type();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextUpdated.chart_type)
  return _s;
}
inline const ::std::string& ContextUpdated::_internal_chart_type() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.chart_type_.Get();
}
inline void ContextUpdated::_internal_set_chart_type(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.chart_type_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::_internal_mutable_chart_type() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.chart_type_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextUpdated::release_chart_type() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextUpdated.chart_type)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000004U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  auto* released = _impl_.chart_type_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.chart_type_.Set("", GetArena());
  }
  return released;
}
inline void ContextUpdated::set_allocated_chart_type(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  _impl_.chart_type_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.chart_type_.IsDefault()) {
    _impl_.chart_type_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextUpdated.chart_type)
}

// string units = 9;
inline void ContextUpdated::clear_units() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.units_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline const ::std::string& ContextUpdated::units() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.units)
  return _internal_units();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextUpdated::set_units(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  _impl_.units_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.units)
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::mutable_units()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  ::std::string* _s = _internal_mutable_units();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextUpdated.units)
  return _s;
}
inline const ::std::string& ContextUpdated::_internal_units() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.units_.Get();
}
inline void ContextUpdated::_internal_set_units(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.units_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::_internal_mutable_units() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.units_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextUpdated::release_units() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextUpdated.units)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000008U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  auto* released = _impl_.units_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.units_.Set("", GetArena());
  }
  return released;
}
inline void ContextUpdated::set_allocated_units(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  }
  _impl_.units_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.units_.IsDefault()) {
    _impl_.units_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextUpdated.units)
}

// string family = 10;
inline void ContextUpdated::clear_family() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.family_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
}
inline const ::std::string& ContextUpdated::family() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextUpdated.family)
  return _internal_family();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextUpdated::set_family(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  _impl_.family_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextUpdated.family)
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::mutable_family()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::std::string* _s = _internal_mutable_family();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextUpdated.family)
  return _s;
}
inline const ::std::string& ContextUpdated::_internal_family() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.family_.Get();
}
inline void ContextUpdated::_internal_set_family(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.family_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextUpdated::_internal_mutable_family() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.family_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextUpdated::release_family() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextUpdated.family)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000010U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  auto* released = _impl_.family_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.family_.Set("", GetArena());
  }
  return released;
}
inline void ContextUpdated::set_allocated_family(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  }
  _impl_.family_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.family_.IsDefault()) {
    _impl_.family_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextUpdated.family)
}

// -------------------------------------------------------------------

// ContextsSnapshot

// repeated .context.v1.ContextUpdated contexts = 1;
inline int ContextsSnapshot::_internal_contexts_size() const {
  return _internal_contexts().size();
}
inline int ContextsSnapshot::contexts_size() const {
  return _internal_contexts_size();
}
inline void ContextsSnapshot::clear_contexts() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.contexts_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::context::v1::ContextUpdated& ContextsSnapshot::contexts(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextsSnapshot.contexts)
  return _internal_contexts().Get(index);
}
inline ::context::v1::ContextUpdated* PROTOBUF_NONNULL ContextsSnapshot::mutable_contexts(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:context.v1.ContextsSnapshot.contexts)
  return _internal_mutable_contexts()->Mutable(index);
}
inline ::context::v1::ContextUpdated* PROTOBUF_NONNULL ContextsSnapshot::add_contexts()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::context::v1::ContextUpdated* _add =
      _internal_mutable_contexts()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:context.v1.ContextsSnapshot.contexts)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>& ContextsSnapshot::contexts() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:context.v1.ContextsSnapshot.contexts)
  return _internal_contexts();
}
inline ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>* PROTOBUF_NONNULL
ContextsSnapshot::mutable_contexts() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:context.v1.ContextsSnapshot.contexts)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_contexts();
}
inline const ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>&
ContextsSnapshot::_internal_contexts() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.contexts_;
}
inline ::google::protobuf::RepeatedPtrField<::context::v1::ContextUpdated>* PROTOBUF_NONNULL
ContextsSnapshot::_internal_mutable_contexts() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.contexts_;
}

// string claim_id = 2;
inline void ContextsSnapshot::clear_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& ContextsSnapshot::claim_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextsSnapshot.claim_id)
  return _internal_claim_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextsSnapshot::set_claim_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.claim_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextsSnapshot.claim_id)
}
inline ::std::string* PROTOBUF_NONNULL ContextsSnapshot::mutable_claim_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_claim_id();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextsSnapshot.claim_id)
  return _s;
}
inline const ::std::string& ContextsSnapshot::_internal_claim_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.claim_id_.Get();
}
inline void ContextsSnapshot::_internal_set_claim_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextsSnapshot::_internal_mutable_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.claim_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextsSnapshot::release_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextsSnapshot.claim_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.claim_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  return released;
}
inline void ContextsSnapshot::set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.claim_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.claim_id_.IsDefault()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextsSnapshot.claim_id)
}

// string node_id = 3;
inline void ContextsSnapshot::clear_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline const ::std::string& ContextsSnapshot::node_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextsSnapshot.node_id)
  return _internal_node_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextsSnapshot::set_node_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  _impl_.node_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextsSnapshot.node_id)
}
inline ::std::string* PROTOBUF_NONNULL ContextsSnapshot::mutable_node_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::std::string* _s = _internal_mutable_node_id();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextsSnapshot.node_id)
  return _s;
}
inline const ::std::string& ContextsSnapshot::_internal_node_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.node_id_.Get();
}
inline void ContextsSnapshot::_internal_set_node_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextsSnapshot::_internal_mutable_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.node_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextsSnapshot::release_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextsSnapshot.node_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000004U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  auto* released = _impl_.node_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.node_id_.Set("", GetArena());
  }
  return released;
}
inline void ContextsSnapshot::set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  _impl_.node_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.node_id_.IsDefault()) {
    _impl_.node_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextsSnapshot.node_id)
}

// uint64 version = 4;
inline void ContextsSnapshot::clear_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline ::uint64_t ContextsSnapshot::version() const {
  // @@protoc_insertion_point(field_get:context.v1.ContextsSnapshot.version)
  return _internal_version();
}
inline void ContextsSnapshot::set_version(::uint64_t value) {
  _internal_set_version(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_set:context.v1.ContextsSnapshot.version)
}
inline ::uint64_t ContextsSnapshot::_internal_version() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.version_;
}
inline void ContextsSnapshot::_internal_set_version(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_ = value;
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace context


// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // proto_2fcontext_2fv1_2fcontext_2eproto_2epb_2eh
